/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual;

import java.util.Objects;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.config.Config;
import org.dussan.vaadin.dvirtual.config.DVirtualConfig;
import org.dussan.vaadin.dvirtual.config.DVirtualSetupConfig;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.info.ApplicationLocale;
import org.dussan.vaadin.dvirtual.info.KeyboardLayout;
import org.dussan.vaadin.dvirtual.info.TranslationLocale;
import org.dussan.vaadin.dvirtual.login.UserLogin;
import org.dussan.vaadin.dvirtual.ui.VirtualUI;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.EncryptionUtil;
import org.dussan.vaadin.dvirtual.utils.EnvironmentUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogErrorUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogInfoUtil;

import com.vaadin.flow.component.AttachEvent;
import com.vaadin.flow.component.Component;
import com.vaadin.flow.component.DetachEvent;
import com.vaadin.flow.component.HasComponents;
import com.vaadin.flow.component.HasSize;
import com.vaadin.flow.component.HasStyle;
import com.vaadin.flow.component.Tag;
import com.vaadin.flow.component.UI;
import com.vaadin.flow.component.dependency.CssImport;
import com.vaadin.flow.component.dependency.JsModule;
import com.vaadin.flow.component.orderedlayout.FlexComponent.Alignment;
import com.vaadin.flow.component.orderedlayout.FlexComponent.JustifyContentMode;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.page.Page;
import com.vaadin.flow.server.VaadinSession;

@Tag( "dvirtual" )
@JsModule( "./icons/brands-iconset-svg.js" )
@JsModule( "./icons/entypo-iconset-svg.js" )
@JsModule( "./icons/feather-iconset-svg.js" )
@JsModule( "./icons/icomoon-iconset-svg.js" )
@JsModule( "./icons/iconic-iconset-svg.js" )
@JsModule( "./icons/jqueryui-iconset-svg.js" )
@JsModule( "./icons/typicons-iconset-svg.js" )
@JsModule( "./icons/wpzoom-iconset-svg.js" )
@JsModule( "./styles/grid-cell-theme.js" )
@JsModule( "./styles/grid-combobox-theme.js" )
@JsModule( "./styles/vaadin-menu-bar-button-theme.js" )
@JsModule( "./styles/vaadin-notification-container-theme.js" )
@JsModule( "./styles/vaadin-tabs-theme.js" )
@CssImport( "./styles/dvirtual.css" )
@CssImport( "./src/jquery/jquery-ui.css" )
@CssImport( "./src/jquery/jquery-ui.structure.css" )
@CssImport( "./src/jquery/jquery-ui.theme.css" )
public class DVirtual
    extends Component
    implements HasComponents, HasSize, HasStyle
{

    private static final long serialVersionUID = 4866822935524336816L;

    private static final int DEFAULT_SESSION_TIMEOUT = 1800; // 1800 seconds = 30 minutes

    private static final int DEFAULT_MOUSE_MOVE_THROTTLE_PERIOD = 200; // miliseconds

    private static final int MAX_INACTIVE_SESSION_INTERVAL = -1;

    private static final String DEFAULT_APPLICATION_CONFIG = "application.xml";

    private static final String DEFAULT_APPLICATIONS_GROUP_CONFIG = "applications_group.xml";

    private static final String DEFAULT_CONFIG_FOLDER = "/etc/dvirtual";

    private static final String DEFAULT_DATABASE_CONFIG = "database.xml";

    private static final String DEFAULT_DVIRTUAL_CONFIG = "dvirtual.xml";

    private static final String DEFAULT_SERVER_CONFIG = "server.xml";

    private static final String DEFAULT_SERVERS_GROUP_CONFIG = "servers_group.xml";

    private static final String DEFAULT_USER_CONFIG = "user.xml";

    private static final String DVIRTUAL_SETUP_CONFIG = "dvirtualsetup.xml";

    private static final String ENV_LOOKUP_CONFIG_FOLDER = "configFolder";

    private static final String ENV_LOOKUP_DVIRTUAL_CONFIG = "dvirtualConfig";

    private static final String JAVASCRIPT_DVIRTUAL = "src/dvirtual.js";

    private static final String JAVASCRIPT_DVIRTUAL_COMMON = "src/dvirtual.common.js";

    private static final String JAVASCRIPT_DVIRTUAL_KEY_CODES = "src/dvirtual.keycodes.js";

    private static final String JAVASCRIPT_JSPDF = "src/jspdf.min.js";

    private static final String JAVASCRIPT_JQUERY = "src/jquery/jquery.min.js";

    private static final String JAVASCRIPT_JQUERY_FULL_SCREEN = "src/jquery/jquery.fullscreen.js";

    private static final String JAVASCRIPT_JQUERY_MOUSE_WHEEL = "src/jquery/jquery.mousewheel.js";

    private static final String JAVASCRIPT_JQUERY_RESIZE_END = "src/jquery/jquery-resizeEnd.js";

    private static final String JAVASCRIPT_JQUERY_UI = "src/jquery/jquery-ui.js";

    private static final String JAVASCRIPT_LZ4 = "src/lz4.min.js";

    private static final String JAVASCRIPT_TASK_TIMER = "src/tasktimer.min.js";

    private static final String STYLE_MINIMAL_HEIGHT = "min-height";

    private static final String STYLE_MINIMAL_HEIGHT_VALUE = "600px";

    private static final String STYLE_MINIMAL_WIDTH = "min-width";

    private static final String STYLE_MINIMAL_WIDTH_VALUE = "800px";

    private boolean allRunningInSameInstance = false;

    private boolean applicationsBarVisible = true;

    private boolean startupNotificationsVisible = true;

    private boolean tooltipsVisible = true;

    private int mouseMoveThrottlePeriod = 0;

    private int sessionTimeout = 0;

    private String applicationConfig = null;

    private String applicationsGroupConfig = null;

    private String configFolder = null;

    private String databaseConfig = null;

    private String dvirtualConfig = null;

    private String loginTitle = null;

    private String serverConfig = null;

    private String serversGroupConfig = null;

    private String userConfig = null;

    private ApplicationLocale applicationLocale = null;

    private KeyboardLayout keyboardLayout = null;

    private TranslationLocale translationLocale = null;

    private UserLogin userLogin = null;

    private VirtualUI virtualUI = null;

    private VirtualUtil virtualUtil = null;

    public DVirtual()
    {
        setSizeFull();
        getStyle().set( STYLE_MINIMAL_WIDTH, STYLE_MINIMAL_WIDTH_VALUE )
                  .set( STYLE_MINIMAL_HEIGHT, STYLE_MINIMAL_HEIGHT_VALUE );
        userLogin = new UserLogin();

        // default settings
        setApplicationLocale( ApplicationLocale.DEFAULT_LOCALE );
        setKeyboardLayout( KeyboardLayout.DEFAULT_LAYOUT );
        setMouseMoveThrottlePeriod( DEFAULT_MOUSE_MOVE_THROTTLE_PERIOD );
        setSessionTimeout( DEFAULT_SESSION_TIMEOUT );
        setConfigFolder( DEFAULT_CONFIG_FOLDER );
        setApplicationConfig( DEFAULT_APPLICATION_CONFIG );
        setApplicationsGroupConfig( DEFAULT_APPLICATIONS_GROUP_CONFIG );
        setDatabaseConfig( DEFAULT_DATABASE_CONFIG );
        setDvirtualConfig( DEFAULT_DVIRTUAL_CONFIG );
        setServerConfig( DEFAULT_SERVER_CONFIG );
        setServersGroupConfig( DEFAULT_SERVERS_GROUP_CONFIG );
        setUserConfig( DEFAULT_USER_CONFIG );

        // load javascript libraries
        Page page = UI.getCurrent()
                      .getPage();
        page.addJavaScript( JAVASCRIPT_JQUERY );
        page.addJavaScript( JAVASCRIPT_JQUERY_RESIZE_END );
        page.addJavaScript( JAVASCRIPT_JQUERY_FULL_SCREEN );
        page.addJavaScript( JAVASCRIPT_JQUERY_MOUSE_WHEEL );
        page.addJavaScript( JAVASCRIPT_JQUERY_UI );
        page.addJavaScript( JAVASCRIPT_JSPDF );
        page.addJavaScript( JAVASCRIPT_TASK_TIMER );
        page.addJavaScript( JAVASCRIPT_DVIRTUAL_COMMON );
        page.addJavaScript( JAVASCRIPT_DVIRTUAL_KEY_CODES );
        // !!!!! DON'T CHANGE ORDER !!!!!
        page.addJavaScript( JAVASCRIPT_LZ4 );
        page.addJavaScript( JAVASCRIPT_DVIRTUAL );

        // read settings from web.xml
        setConfigFolder( EnvironmentUtil.readStringEnvEntry( ENV_LOOKUP_CONFIG_FOLDER ) );
        setDvirtualConfig( EnvironmentUtil.readStringEnvEntry( ENV_LOOKUP_DVIRTUAL_CONFIG ) );
    }

    private DVirtual getThis()
    {
        return this;
    }

    private void init()
    {
        // new VirtualUI instance
        virtualUI = new VirtualUI()
        {
            private static final long serialVersionUID = 2646015316186837005L;

            @Override
            public void logOut()
            {
                if ( isUserLoginFormEnabled() )
                {
                    Optional<UI> uiOptional = getThis().getUI();
                    uiOptional.ifPresent( ui -> ui.getSession()
                                                  .lock() );
                    uiOptional.ifPresent( ui -> ui.access( () -> {
                        userLogin = new UserLogin();
                        userLogin.setTranslationLocale( getTranslationLocale() );
                        userLogin.setTitle( getLoginTitle() );
                        userLogin.setSignIn( false );
                        userLogin.addLoginListener( event -> {
                            userLogin.setUsername( StringUtils.EMPTY );
                            userLogin.setPassword( StringUtils.EMPTY );
                            virtualUtil.setSetupFileDeleted( true );
                            virtualUtil.setSetupUserActive( false );
                            virtualUI.setUserData( event.getUsername(), event.getPassword() );
                            if ( !userLogin.isSignIn() && Objects.nonNull( virtualUI.getUserData() ) )
                            {
                                if ( virtualUI.getUserData()
                                              .isAdministrator()
                                              .booleanValue() )
                                {
                                    virtualUtil.setSetupFileDeleted( !( new DVirtualSetupConfig( configFolder, DVIRTUAL_SETUP_CONFIG ).isInitialized() ) );
                                }
                                virtualUtil.setDefaultValues();
                                userLogin.setError( false );
                                userLogin.setSignIn( true );
                                getThis().removeAll();
                                getThis().add( virtualUI.getLayout() );
                            }
                            else
                            {
                                DVirtualSetupConfig dVirtualSetupConfig = new DVirtualSetupConfig( configFolder, DVIRTUAL_SETUP_CONFIG );
                                if ( dVirtualSetupConfig.isInitialized() && StringUtils.equals( event.getUsername(), dVirtualSetupConfig.getUsername() )
                                    && StringUtils.equals( event.getPassword(), dVirtualSetupConfig.getPassword() ) )
                                {
                                    // dVirtual setup user
                                    virtualUtil.setDefaultValues();
                                    userLogin.setError( false );
                                    userLogin.setEnabled( true );
                                    userLogin.setSignIn( true );
                                    virtualUtil.setSetupUserActive( true );
                                    virtualUtil.setUserName( event.getUsername() );
                                    virtualUtil.setUserPassword( EncryptionUtil.encrypt( event.getPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ) );
                                    getThis().removeAll();
                                    getThis().add( virtualUI.getLayout() );
                                }
                                else
                                {
                                    virtualUtil.setUserName( null );
                                    virtualUtil.setUserPassword( null );
                                    userLogin.setError( true );
                                }
                                dVirtualSetupConfig.reset();
                            }
                        } );
                        stopAllServices();

                        VerticalLayout userLoginLayout = new VerticalLayout( userLogin );
                        userLoginLayout.setAlignItems( Alignment.CENTER );
                        userLoginLayout.setJustifyContentMode( JustifyContentMode.CENTER );
                        userLoginLayout.setMargin( false );
                        userLoginLayout.setPadding( false );
                        userLoginLayout.setSpacing( false );
                        userLoginLayout.setSizeFull();

                        getThis().removeAll();
                        getThis().add( userLoginLayout );
                        ui.push();
                    } ) );
                    uiOptional.ifPresent( ui -> ui.getSession()
                                                  .unlock() );
                }
                else
                {
                    stopAllServices();
                    userLogin.setSignIn( false );
                    virtualUI.removeAll();
                    removeAll();
                    getThis().getElement()
                             .removeFromParent();
                }
                getVirtualUtil().setUserName( null );
                getVirtualUtil().setUserPassword( null );
            }
        };

        // dVirtual config
        if ( StringUtils.isNotBlank( getConfigFolder() ) && StringUtils.isNotBlank( getDvirtualConfig() ) )
        {
            DVirtualConfig dVirtualConfig = new DVirtualConfig( getConfigFolder(), getDvirtualConfig() );
            setMouseMoveThrottlePeriod( dVirtualConfig.getMouseMoveThrottlePeriod() );
            setSessionTimeout( dVirtualConfig.getSessionTimeout() );
            setKeyboardLayout( dVirtualConfig.getKeyboardLayout() );
            setApplicationLocale( dVirtualConfig.getApplicationLocale() );
            setTranslationLocale( dVirtualConfig.getTranslationLocale() );
            setLoginTitle( dVirtualConfig.getLoginTitle() );
            setAllRunningInSameInstance( dVirtualConfig.isAllRunningInSameInstance() );
            setApplicationsBarVisible( dVirtualConfig.isApplicationsBarVisible() );
            setStartupNotificationsVisible( dVirtualConfig.isStartupNotificationsVisible() );
            setTooltipsVisible( dVirtualConfig.isTooltipsVisible() );
            setDatabaseConfig( dVirtualConfig.getDatabaseConfig() );
            setApplicationConfig( dVirtualConfig.getApplicationConfig() );
            setApplicationsGroupConfig( dVirtualConfig.getApplicationsGroupConfig() );
            setServerConfig( dVirtualConfig.getServerConfig() );
            setServersGroupConfig( dVirtualConfig.getServersGroupConfig() );
            setUserConfig( dVirtualConfig.getUserConfig() );

            if ( !dVirtualConfig.isInitialized() )
            {
                new DialogErrorUtil( DialogErrorUtil.ERROR_DVIRTUAL ).open();
                stopAllServices();
                return;
            }

            // virtualUtil
            virtualUtil = new VirtualUtil()
            {
                private static final long serialVersionUID = 1587656168684238070L;

                @Override
                public void logOutUser()
                {
                    if ( !isUserLogOut() )
                    {
                        setUserLogOut( true );
                        virtualUI.logOut();
                    }
                }
            };
            virtualUtil.setMouseMoveThrottlePeriod( getMouseMoveThrottlePeriod() );
            virtualUtil.setTranslationLocale( getTranslationLocale() );
            virtualUtil.setSecureIV( dVirtualConfig.getSaltIV() );
            virtualUtil.setSecretKeySpec( dVirtualConfig.getSalt(), dVirtualConfig.getSaltKey() );
            virtualUI.setVirtualUtil( virtualUtil );
        }
        else
        {
            new DialogErrorUtil( DialogErrorUtil.ERROR_DVIRTUAL ).open();
            stopAllServices();
            return;
        }

        // database initialization
        if ( StringUtils.isNotBlank( getConfigFolder() ) && StringUtils.isNotBlank( getDatabaseConfig() ) )
        {
            Database database = new Database( getConfigFolder(), getDatabaseConfig() );
            if ( database.isNotInitialized() )
            {
                new DialogErrorUtil( DialogErrorUtil.ERROR_DVIRTUAL ).open();
                stopAllServices();
                return;
            }
            virtualUI.setDatabase( database );
        }
        else
        {
            new DialogErrorUtil( DialogErrorUtil.ERROR_DVIRTUAL ).open();
            stopAllServices();
            return;
        }

        // import data: servers, server groups, applications, application groups, users
        Config config = new Config( getConfigFolder(), getServerConfig(), getServersGroupConfig(), getApplicationConfig(), getApplicationsGroupConfig(), getUserConfig() );
        config.importData( virtualUI.getDatabase() );

        // user data initialization
        userLogin.setTranslationLocale( getTranslationLocale() );
        userLogin.setTitle( getLoginTitle() );
        if ( userLogin.isLoginEnabled() )
        {
            userLogin.setError( false );
            userLogin.setSignIn( false );
            userLogin.addLoginListener( event -> {
                userLogin.setUsername( StringUtils.EMPTY );
                userLogin.setPassword( StringUtils.EMPTY );
                virtualUtil.setSetupFileDeleted( true );
                virtualUtil.setSetupUserActive( false );
                virtualUI.setUserData( event.getUsername(), event.getPassword() );
                if ( !userLogin.isSignIn() && Objects.nonNull( virtualUI.getUserData() ) )
                {
                    if ( virtualUI.getUserData()
                                  .isAdministrator()
                                  .booleanValue() )
                    {
                        virtualUtil.setSetupFileDeleted( !( new DVirtualSetupConfig( configFolder, DVIRTUAL_SETUP_CONFIG ).isInitialized() ) );
                    }
                    userLogin.setError( false );
                    userLogin.setEnabled( true );
                    userLogin.setSignIn( true );
                    getThis().removeAll();
                    getThis().add( virtualUI.getLayout() );
                }
                else
                {
                    DVirtualSetupConfig dVirtualSetupConfig = new DVirtualSetupConfig( configFolder, DVIRTUAL_SETUP_CONFIG );
                    if ( dVirtualSetupConfig.isInitialized() && StringUtils.equals( event.getUsername(), dVirtualSetupConfig.getUsername() )
                        && StringUtils.equals( event.getPassword(), dVirtualSetupConfig.getPassword() ) )
                    {
                        // dVirtual setup user
                        userLogin.setError( false );
                        userLogin.setEnabled( true );
                        userLogin.setSignIn( true );
                        virtualUtil.setSetupFileDeleted( false );
                        virtualUtil.setSetupUserActive( true );
                        virtualUtil.setUserName( event.getUsername() );
                        virtualUtil.setUserPassword( EncryptionUtil.encrypt( event.getPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ) );
                        getThis().removeAll();
                        getThis().add( virtualUI.getLayout() );
                    }
                    else
                    {
                        userLogin.setError( true );
                    }
                    dVirtualSetupConfig.reset();
                }
            } );

            VerticalLayout userLoginLayout = new VerticalLayout( userLogin );
            userLoginLayout.setAlignItems( Alignment.CENTER );
            userLoginLayout.setJustifyContentMode( JustifyContentMode.CENTER );
            userLoginLayout.setMargin( false );
            userLoginLayout.setPadding( false );
            userLoginLayout.setSpacing( false );
            userLoginLayout.setSizeFull();
            removeAll();
            add( userLoginLayout );
        }
        else
        {
            virtualUI.setUserData( userLogin.getUsername(), userLogin.getPassword() );
            if ( Objects.nonNull( virtualUI.getUserData() ) )
            {
                userLogin.setEnabled( true );
                userLogin.setSignIn( true );
                add( virtualUI.getLayout() );
            }
            else
            {
                userLogin.setSignIn( false );
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.LOGIN_ERROR_MESSAGE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
        }
        virtualUtil.setMouseMoveThrottlePeriod( getMouseMoveThrottlePeriod() );
        virtualUtil.setTranslationLocale( getTranslationLocale() );
        virtualUtil.setApplicationLocale( getApplicationLocale() );
        virtualUtil.setKeyboardLayout( getKeyboardLayout() );
        virtualUtil.setAllRunningInSameInstance( isAllRunningInSameInstance() );
        virtualUtil.setApplicationsBarVisible( isApplicationsBarVisible() );
        virtualUtil.setStartupNotificationsVisible( isStartupNotificationsVisible() );
        virtualUI.setSessionTimeout( getSessionTimeout() );

        // session timeout: disabled, it is controlled by dVirtual util
        VaadinSession.getCurrent()
                     .getSession()
                     .setMaxInactiveInterval( MAX_INACTIVE_SESSION_INTERVAL );
    }

    private void stopAllServices()
    {
        if ( Objects.nonNull( virtualUI ) && Objects.nonNull( virtualUI.getVirtualUtil() ) )
        {
            virtualUI.getVirtualUtil()
                     .stopAllStartedServices();
            virtualUI.getDatabase()
                     .close();
        }
    }

    public boolean isUserSignIn()
    {
        return userLogin.isSignIn();
    }

    public int getMouseMoveThrottlePeriod()
    {
        return mouseMoveThrottlePeriod;
    }

    public void setMouseMoveThrottlePeriod( Integer mouseMoveThrottlePeriod )
    {
        this.mouseMoveThrottlePeriod = Objects.isNull( mouseMoveThrottlePeriod ) || mouseMoveThrottlePeriod <= 0 ? DEFAULT_MOUSE_MOVE_THROTTLE_PERIOD : mouseMoveThrottlePeriod;
    }

    public int getSessionTimeout()
    {
        return sessionTimeout;
    }

    public void setSessionTimeout( Integer sessionTimeout )
    {
        this.sessionTimeout = Objects.isNull( sessionTimeout ) || sessionTimeout <= 0 ? DEFAULT_SESSION_TIMEOUT : sessionTimeout;
        this.sessionTimeout *= 1000;
    }

    public ApplicationLocale getApplicationLocale()
    {
        return applicationLocale;
    }

    public void setApplicationLocale( ApplicationLocale applicationLocale )
    {
        if ( Objects.nonNull( applicationLocale ) )
        {
            this.applicationLocale = applicationLocale;
        }
    }

    public void setApplicationLocale( String applicationLocale )
    {
        if ( StringUtils.isNotBlank( applicationLocale ) )
        {
            this.applicationLocale = ApplicationLocale.getApplicationLocale( applicationLocale );
        }
    }

    public TranslationLocale getTranslationLocale()
    {
        return translationLocale;
    }

    public void setTranslationLocale( TranslationLocale translationLocale )
    {
        if ( Objects.nonNull( translationLocale ) )
        {
            this.translationLocale = translationLocale;
        }

        if ( Objects.nonNull( userLogin ) )
        {
            userLogin.setTranslationLocale( this.translationLocale );
        }

        if ( Objects.nonNull( virtualUI ) && Objects.nonNull( virtualUI.getVirtualUtil() ) )
        {
            virtualUI.getVirtualUtil()
                     .setTranslationLocale( this.translationLocale );
        }
    }

    public void setTranslationLocale( String translationLocale )
    {
        if ( StringUtils.isNotBlank( translationLocale ) )
        {
            setTranslationLocale( TranslationLocale.getTranslationLocale( translationLocale ) );
        }
    }

    public KeyboardLayout getKeyboardLayout()
    {
        return keyboardLayout;
    }

    public void setKeyboardLayout( KeyboardLayout keyboardLayout )
    {
        if ( Objects.nonNull( keyboardLayout ) )
        {
            this.keyboardLayout = keyboardLayout;
        }
    }

    public void setKeyboardLayout( String keyboardLayout )
    {
        if ( StringUtils.isNotBlank( keyboardLayout ) )
        {
            this.keyboardLayout = KeyboardLayout.getKeyboardLayout( keyboardLayout );
        }
    }

    public String getLoginTitle()
    {
        return loginTitle;
    }

    public void setLoginTitle( String loginTitle )
    {
        if ( StringUtils.isNotBlank( loginTitle ) )
        {
            this.loginTitle = loginTitle;
        }
    }

    public String getConfigFolder()
    {
        return configFolder;
    }

    public void setConfigFolder( String configFolder )
    {
        if ( StringUtils.isNotBlank( configFolder ) )
        {
            this.configFolder = configFolder;
        }
    }

    public String getApplicationConfig()
    {
        return applicationConfig;
    }

    public void setApplicationConfig( String applicationConfig )
    {
        if ( StringUtils.isNotBlank( applicationConfig ) )
        {
            this.applicationConfig = applicationConfig;
        }
    }

    public String getApplicationsGroupConfig()
    {
        return applicationsGroupConfig;
    }

    public void setApplicationsGroupConfig( String applicationsGroupConfig )
    {
        if ( StringUtils.isNotBlank( applicationsGroupConfig ) )
        {
            this.applicationsGroupConfig = applicationsGroupConfig;
        }
    }

    public String getDatabaseConfig()
    {
        return databaseConfig;
    }

    public void setDatabaseConfig( String databaseConfig )
    {
        if ( StringUtils.isNotBlank( databaseConfig ) )
        {
            this.databaseConfig = databaseConfig;
        }
    }

    public String getDvirtualConfig()
    {
        return dvirtualConfig;
    }

    public void setDvirtualConfig( String dvirtualConfig )
    {
        if ( StringUtils.isNotBlank( dvirtualConfig ) )
        {
            this.dvirtualConfig = dvirtualConfig;
        }
    }

    public String getServerConfig()
    {
        return serverConfig;
    }

    public void setServerConfig( String serverConfig )
    {
        if ( StringUtils.isNotBlank( serverConfig ) )
        {
            this.serverConfig = serverConfig;
        }
    }

    public String getServersGroupConfig()
    {
        return serversGroupConfig;
    }

    public void setServersGroupConfig( String serversGroupConfig )
    {
        if ( StringUtils.isNotBlank( serversGroupConfig ) )
        {
            this.serversGroupConfig = serversGroupConfig;
        }
    }

    public String getUserConfig()
    {
        return userConfig;
    }

    public void setUserConfig( String userConfig )
    {
        if ( StringUtils.isNotBlank( userConfig ) )
        {
            this.userConfig = userConfig;
        }
    }

    public boolean isUserLoginFormEnabled()
    {
        return userLogin.isLoginEnabled();
    }

    public void setUserLoginFormEnabled( boolean loginEnabled )
    {
        userLogin.setLoginEnabled( loginEnabled );
    }

    public void setUserLoginData( String userName, String userPassword )
    {
        boolean loginEnabled = userLogin.isLoginEnabled();
        userLogin = new UserLogin( userName, userPassword );
        userLogin.setTranslationLocale( getTranslationLocale() );
        userLogin.setTitle( getLoginTitle() );
        userLogin.setLoginEnabled( loginEnabled );
    }

    public void setUserLoginFormTitle( String title )
    {
        userLogin.setTitle( title );
    }

    public boolean isAllRunningInSameInstance()
    {
        return allRunningInSameInstance;
    }

    public void setAllRunningInSameInstance( Boolean allRunningInSameInstance )
    {
        if ( Objects.nonNull( allRunningInSameInstance ) )
        {
            this.allRunningInSameInstance = allRunningInSameInstance;
            if ( Objects.nonNull( virtualUI ) )
            {
                virtualUtil.setAllRunningInSameInstance( isAllRunningInSameInstance() );
            }
        }
    }

    public boolean isApplicationsBarVisible()
    {
        return applicationsBarVisible;
    }

    public void setApplicationsBarVisible( Boolean applicationsBarVisible )
    {
        if ( Objects.nonNull( applicationsBarVisible ) )
        {
            this.applicationsBarVisible = applicationsBarVisible;
            if ( Objects.nonNull( virtualUtil ) )
            {
                virtualUtil.setApplicationsBarVisible( isApplicationsBarVisible() );
            }
        }
    }

    public boolean isStartupNotificationsVisible()
    {
        return startupNotificationsVisible;
    }

    public void setStartupNotificationsVisible( Boolean startupNotificationsVisible )
    {
        if ( Objects.nonNull( startupNotificationsVisible ) )
        {
            this.startupNotificationsVisible = startupNotificationsVisible;
            if ( Objects.nonNull( virtualUtil ) )
            {
                virtualUtil.setStartupNotificationsVisible( startupNotificationsVisible );
            }
        }
    }

    public boolean isTooltipsVisible()
    {
        return tooltipsVisible;
    }

    public void setTooltipsVisible( Boolean tooltipsVisible )
    {
        if ( Objects.nonNull( tooltipsVisible ) )
        {
            this.tooltipsVisible = tooltipsVisible;
            if ( Objects.nonNull( virtualUtil ) )
            {
                virtualUtil.setTooltipsVisible( tooltipsVisible );
            }
        }
    }

    @Override
    protected void onAttach( AttachEvent attachEvent )
    {
        init();
    }

    @Override
    protected void onDetach( DetachEvent detachEvent )
    {
        stopAllServices();
        userLogin.setSignIn( false );
        virtualUI.removeAll();
        if ( Objects.nonNull( virtualUtil ) )
        {
            virtualUtil.setUserName( null );
            virtualUtil.setUserPassword( null );
        }
    }

}
