/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.config;

import java.io.Serializable;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.ApplicationsGroupData;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.database.data.ServersGroupData;
import org.dussan.vaadin.dvirtual.database.data.UserData;

public class Config
    implements Serializable
{

    private static final long serialVersionUID = 2798141748704099460L;

    private static final int NO_ELEMENTS = 0;

    private String configApplication = null;

    private String configApplicationsGroup = null;

    private String configFolder = null;

    private String configUser = null;

    private String configServer = null;

    private String configServersGroup = null;

    public Config( String configFolder, String configServer, String configServersGroup, String configApplication, String configApplicationsGroup, String configUser )
    {
        this.configFolder = configFolder;
        this.configServer = configServer;
        this.configServersGroup = configServersGroup;
        this.configApplication = configApplication;
        this.configApplicationsGroup = configApplicationsGroup;
        this.configUser = configUser;
    }

    private boolean fileExists( String configFolder, String configFile )
    {
        return Paths.get( configFolder, configFile )
                    .toFile()
                    .exists();
    }

    public void importData( Database database )
    {
        // only import new data when users table is still empty
        if ( StringUtils.isNotBlank( configFolder ) && Objects.nonNull( database ) && database.isInitialized() && database.getUsersRows( null ) == NO_ELEMENTS )
        {
            // import server data
            if ( StringUtils.isNotBlank( configServer ) && fileExists( configFolder, configServer ) )
            {
                ServerConfig serverConfig = new ServerConfig( configFolder, configServer );
                if ( CollectionUtils.isNotEmpty( serverConfig.getServerDatas() ) )
                {
                    serverConfig.getServerDatas()
                                .stream()
                                .filter( Objects::nonNull )
                                .forEach( serverData -> database.addServer( serverData ) );
                }
            }

            // import servers group data
            if ( StringUtils.isNotBlank( configServersGroup ) && fileExists( configFolder, configServersGroup ) )
            {
                ServersGroupConfig serverGroupConfig = new ServersGroupConfig( configFolder, configServersGroup );
                if ( CollectionUtils.isNotEmpty( serverGroupConfig.getServersGroupDatas() ) )
                {
                    serverGroupConfig.getServersGroupDatas()
                                     .stream()
                                     .filter( Objects::nonNull )
                                     .forEach( serversGroupData -> {
                                         List<Long> ids = new ArrayList<>();
                                         List<String> names = new ArrayList<>();

                                         serversGroupData.getServerDatas()
                                                         .stream()
                                                         .distinct()
                                                         .forEach( serverData -> {
                                                             if ( Objects.nonNull( serverData.getId() ) && !Objects.equals( serverData.getId(), ServerData.NEW_SERVER ) )
                                                             {
                                                                 ids.add( serverData.getId() );
                                                             }
                                                             else if ( StringUtils.isNotBlank( StringUtils.stripToNull( serverData.getName() ) ) )
                                                             {
                                                                 names.add( serverData.getName() );
                                                             }
                                                         } );

                                         List<ServerData> serverDatas = new ArrayList<>();
                                         if ( CollectionUtils.isNotEmpty( ids ) )
                                         {
                                             serverDatas.addAll( database.getServers( ids.toArray( Long[]::new ) ) );
                                         }
                                         if ( CollectionUtils.isNotEmpty( names ) )
                                         {
                                             serverDatas.addAll( database.getServers( names.toArray( String[]::new ) ) );
                                         }
                                         serverDatas = serverDatas.stream()
                                                                  .distinct()
                                                                  .collect( Collectors.toList() );
                                         serversGroupData.setServerDatas( serverDatas );
                                         database.addServersGroup( serversGroupData );
                                     } );
                }
            }

            // import application data
            if ( StringUtils.isNotBlank( configApplication ) && fileExists( configFolder, configApplication ) )
            {
                ApplicationConfig applicationConfig = new ApplicationConfig( configFolder, configApplication );
                if ( CollectionUtils.isNotEmpty( applicationConfig.getApplicationDatas() ) )
                {
                    applicationConfig.getApplicationDatas()
                                     .stream()
                                     .filter( Objects::nonNull )
                                     .forEach( applicationData -> {
                                         List<Long> ids = new ArrayList<>();
                                         List<String> names = new ArrayList<>();

                                         applicationData.getServersGroupDatas()
                                                        .stream()
                                                        .distinct()
                                                        .forEach( serversGroupData -> {
                                                            if ( Objects.nonNull( serversGroupData.getId() ) && !Objects.equals( serversGroupData.getId(), ServersGroupData.NEW_SERVERS_GROUP ) )
                                                            {
                                                                ids.add( serversGroupData.getId() );
                                                            }
                                                            else if ( StringUtils.isNotBlank( StringUtils.stripToNull( serversGroupData.getName() ) ) )
                                                            {
                                                                names.add( serversGroupData.getName() );
                                                            }
                                                        } );

                                         List<ServersGroupData> serversGroupDatas = new ArrayList<>();
                                         if ( CollectionUtils.isNotEmpty( ids ) )
                                         {
                                             serversGroupDatas.addAll( database.getServerGroups( ids.toArray( Long[]::new ) ) );
                                         }
                                         if ( CollectionUtils.isNotEmpty( names ) )
                                         {
                                             serversGroupDatas.addAll( database.getServerGroups( names.toArray( String[]::new ) ) );
                                         }
                                         serversGroupDatas = serversGroupDatas.stream()
                                                                              .distinct()
                                                                              .collect( Collectors.toList() );
                                         applicationData.setServersGroupDatas( serversGroupDatas );
                                         database.addApplication( applicationData );
                                     } );
                }
            }

            // import applications group data
            if ( StringUtils.isNotBlank( configApplicationsGroup ) && fileExists( configFolder, configApplicationsGroup ) )
            {
                ApplicationsGroupConfig applicationsGroupConfig = new ApplicationsGroupConfig( configFolder, configApplicationsGroup );
                if ( CollectionUtils.isNotEmpty( applicationsGroupConfig.getApplicationsGroupDatas() ) )
                {
                    applicationsGroupConfig.getApplicationsGroupDatas()
                                           .stream()
                                           .filter( Objects::nonNull )
                                           .forEach( applicationsGroupData -> {
                                               List<Long> ids = new ArrayList<>();
                                               List<String> names = new ArrayList<>();

                                               applicationsGroupData.getApplicationDatas()
                                                                    .stream()
                                                                    .distinct()
                                                                    .forEach( applicationData -> {
                                                                        if ( Objects.nonNull( applicationData.getId() ) && !Objects.equals( applicationData.getId(), ApplicationData.NEW_APPLICATION ) )
                                                                        {
                                                                            ids.add( applicationData.getId() );
                                                                        }
                                                                        else if ( StringUtils.isNotBlank( StringUtils.stripToNull( applicationData.getName() ) ) )
                                                                        {
                                                                            names.add( applicationData.getName() );
                                                                        }
                                                                    } );

                                               List<ApplicationData> applicationDatas = new ArrayList<>();
                                               if ( CollectionUtils.isNotEmpty( ids ) )
                                               {
                                                   applicationDatas.addAll( database.getApplications( ids.toArray( Long[]::new ) ) );
                                               }
                                               if ( CollectionUtils.isNotEmpty( names ) )
                                               {
                                                   applicationDatas.addAll( database.getApplications( names.toArray( String[]::new ) ) );
                                               }
                                               applicationDatas = applicationDatas.stream()
                                                                                  .distinct()
                                                                                  .collect( Collectors.toList() );
                                               applicationsGroupData.setApplicationDatas( applicationDatas );
                                               database.addApplicationsGroup( applicationsGroupData );
                                           } );
                }
            }

            // import user data
            if ( StringUtils.isNotBlank( configUser ) && fileExists( configFolder, configUser ) )
            {
                UserConfig userConfig = new UserConfig( configFolder, configUser );
                if ( CollectionUtils.isNotEmpty( userConfig.getUserDatas() ) )
                {
                    Map<Long, ApplicationData> applicationWithIdDatas = new HashMap<>();
                    Map<String, ApplicationData> applicationWithNameDatas = new HashMap<>();
                    Map<Long, ApplicationsGroupData> applicationsGroupWithIdDatas = new HashMap<>();
                    Map<String, ApplicationsGroupData> applicationsGroupWithNameDatas = new HashMap<>();
                    MapUtils.populateMap( applicationWithIdDatas, database.getAllApplications(), ApplicationData::getId );
                    MapUtils.populateMap( applicationWithNameDatas, database.getAllApplications(), ApplicationData::getName );
                    MapUtils.populateMap( applicationsGroupWithIdDatas, database.getAllApplicationGroups(), ApplicationsGroupData::getId );
                    MapUtils.populateMap( applicationsGroupWithNameDatas, database.getAllApplicationGroups(), ApplicationsGroupData::getName );

                    userConfig.getUserDatas()
                              .stream()
                              .filter( Objects::nonNull )
                              .forEach( userData -> {
                                  long userId = database.addUser( userData );
                                  if ( userId != UserData.UNDEFINED_USER )
                                  {
                                      userConfig.getUserApplicationDatas()
                                                .stream()
                                                .filter( userApplicationData -> StringUtils.equals( userApplicationData.getUserData()
                                                                                                                       .getUserName(),
                                                                                                    userData.getUserName() ) )
                                                .forEach( userApplicationData -> {
                                                    ApplicationData applicationData = userApplicationData.getApplicationData();
                                                    if ( applicationWithIdDatas.containsKey( applicationData.getId() ) )
                                                    {
                                                        database.addUserApplication( userData, applicationWithIdDatas.get( applicationData.getId() ) );
                                                    }
                                                    else if ( applicationWithNameDatas.containsKey( applicationData.getName() ) )
                                                    {
                                                        database.addUserApplication( userData, applicationWithNameDatas.get( applicationData.getName() ) );
                                                    }
                                                } );
                                      userConfig.getUserApplicationsGroupDatas()
                                                .stream()
                                                .filter( userApplicationsGroupData -> StringUtils.equals( userApplicationsGroupData.getUserData()
                                                                                                                                   .getUserName(),
                                                                                                          userData.getUserName() ) )
                                                .forEach( userApplicationsGroupData -> {
                                                    ApplicationsGroupData applicationsGroupData = userApplicationsGroupData.getApplicationsGroupData();
                                                    if ( applicationsGroupWithIdDatas.containsKey( applicationsGroupData.getId() ) )
                                                    {
                                                        database.addUserApplicationsGroup( userData, applicationsGroupWithIdDatas.get( applicationsGroupData.getId() ) );
                                                    }
                                                    else if ( applicationsGroupWithNameDatas.containsKey( applicationsGroupData.getName() ) )
                                                    {
                                                        database.addUserApplicationsGroup( userData, applicationsGroupWithNameDatas.get( applicationsGroupData.getName() ) );
                                                    }
                                                } );
                                  }
                              } );
                }
            }
        }
    }

}
