/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.database.data;

import java.io.Serializable;
import java.util.List;

import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.JoinColumn;
import javax.persistence.JoinTable;
import javax.persistence.ManyToMany;
import javax.persistence.ManyToOne;
import javax.persistence.Table;

import org.apache.commons.lang3.StringUtils;

@Entity( name = "ApplicationData" )
@Table( name = "dv_application" )
public class ApplicationData
    implements Serializable
{

    private static final long serialVersionUID = 8205623811038463763L;

    public static final long NEW_APPLICATION = 0;

    public static final long UNDEFINED_APPLICATION = -1;

    public static final Boolean DEFAULT_ENABLED = Boolean.TRUE;

    public static final Boolean DEFAULT_SAME_INSTANCE = Boolean.FALSE;

    public static final Boolean DEFAULT_SERVER_AUTO_SELECT = Boolean.TRUE;

    public static final Integer DEFAULT_ONE_INSTANCE = 0;

    public static final Integer ONE_INSTANCE_DISABLED = 3;

    public static final Integer ONE_INSTANCE_ENABLED = 2;

    public static final Integer ONE_INSTANCE_ENABLED_LOCKED = 1;

    public static final String NEW_DATA = StringUtils.EMPTY;

    @Id
    @GeneratedValue( strategy = GenerationType.AUTO )
    private Long id;

    @Column( name = "name", length = 100, nullable = false )
    private String name;

    @Column( name = "command", length = 200, nullable = false )
    private String command;

    @Column( name = "command_options", length = 200, nullable = true )
    private String commandOptions;

    @Column( name = "note", length = 200, nullable = true )
    private String note;

    @Column( name = "server_auto_select", nullable = false )
    private Boolean serverAutoSelect;

    @Column( name = "one_instance", nullable = false )
    private Integer oneInstance;

    @Column( name = "same_instance", nullable = false )
    private Boolean sameInstance;

    @Column( name = "enabled", nullable = false )
    private Boolean enabled;

    @ManyToOne
    @JoinColumn( name = "icon_id", referencedColumnName = "id", nullable = true )
    private IconData iconData;

    @ManyToMany( fetch = FetchType.EAGER )
    @JoinTable( name = "dv_application_servers_group", //
                    joinColumns = { @JoinColumn( name = "application_id", referencedColumnName = "id" ) }, //
                    inverseJoinColumns = { @JoinColumn( name = "servers_group_id", referencedColumnName = "id" ) } )
    private List<ServersGroupData> serversGroupDatas;

    public ApplicationData()
    {
        setServerAutoSelect( DEFAULT_SERVER_AUTO_SELECT );
        setOneInstance( DEFAULT_ONE_INSTANCE );
        setSameInstance( DEFAULT_SAME_INSTANCE );
        setEnabled( DEFAULT_ENABLED );
    }

    public ApplicationData( long id, String name, String command, String commandOptions, String note, boolean serverAutoSelect, int oneInstance, boolean sameInstance, boolean enabled,
                            IconData iconData, List<ServersGroupData> serversGroupDatas )
    {
        setId( id );
        setName( name );
        setCommand( command );
        setCommandOptions( commandOptions );
        setNote( note );
        setServerAutoSelect( serverAutoSelect );
        setOneInstance( oneInstance );
        setSameInstance( sameInstance );
        setEnabled( enabled );
        setIconData( iconData );
        setServersGroupDatas( serversGroupDatas );
    }

    public Long getId()
    {
        return id;
    }

    public void setId( Long id )
    {
        this.id = id;
    }

    public String getName()
    {
        return name;
    }

    public void setName( String name )
    {
        this.name = StringUtils.stripToEmpty( name );
    }

    public String getCommand()
    {
        return command;
    }

    public void setCommand( String command )
    {
        this.command = StringUtils.stripToEmpty( command );
    }

    public String getCommandOptions()
    {
        return commandOptions;
    }

    public void setCommandOptions( String commandOptions )
    {
        this.commandOptions = commandOptions;
    }

    public String getNote()
    {
        return note;
    }

    public void setNote( String note )
    {
        this.note = StringUtils.stripToEmpty( note );
    }

    public Boolean isServerAutoSelect()
    {
        return serverAutoSelect;
    }

    public void setServerAutoSelect( Boolean serverAutoSelect )
    {
        this.serverAutoSelect = serverAutoSelect;
    }

    public Integer getOneInstance()
    {
        return oneInstance;
    }

    public Boolean isOneInstance()
    {
        return ( getOneInstance() < ApplicationData.ONE_INSTANCE_DISABLED );
    }

    public Boolean isNotOneInstance()
    {
        return !isOneInstance();
    }

    public void setOneInstance( Integer oneInstance )
    {
        this.oneInstance = oneInstance;
    }

    public Boolean isSameInstance()
    {
        return sameInstance;
    }

    public Boolean isNotSameInstance()
    {
        return !sameInstance;
    }

    public void setSameInstance( Boolean sameInstance )
    {
        this.sameInstance = sameInstance;
    }

    public Boolean isEnabled()
    {
        return enabled;
    }

    public void setEnabled( Boolean enabled )
    {
        this.enabled = enabled;
    }

    public IconData getIconData()
    {
        return iconData;
    }

    public void setIconData( IconData iconData )
    {
        this.iconData = iconData;
    }

    public List<ServersGroupData> getServersGroupDatas()
    {
        return serversGroupDatas;
    }

    public void setServersGroupDatas( List<ServersGroupData> serversGroupDatas )
    {
        this.serversGroupDatas = serversGroupDatas;
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + ( ( command == null ) ? 0 : command.hashCode() );
        result = prime * result + ( ( commandOptions == null ) ? 0 : commandOptions.hashCode() );
        result = prime * result + ( ( enabled == null ) ? 0 : enabled.hashCode() );
        result = prime * result + ( ( id == null ) ? 0 : id.hashCode() );
        result = prime * result + ( ( name == null ) ? 0 : name.hashCode() );
        result = prime * result + ( ( note == null ) ? 0 : note.hashCode() );
        result = prime * result + ( ( oneInstance == null ) ? 0 : oneInstance.hashCode() );
        result = prime * result + ( ( sameInstance == null ) ? 0 : sameInstance.hashCode() );
        result = prime * result + ( ( serverAutoSelect == null ) ? 0 : serverAutoSelect.hashCode() );
        return result;
    }

    @Override
    public boolean equals( Object obj )
    {
        if ( this == obj )
        {
            return true;
        }
        if ( obj == null )
        {
            return false;
        }
        if ( !( obj instanceof ApplicationData ) )
        {
            return false;
        }
        ApplicationData other = (ApplicationData) obj;
        if ( command == null )
        {
            if ( other.command != null )
            {
                return false;
            }
        }
        else if ( !command.equals( other.command ) )
        {
            return false;
        }
        if ( commandOptions == null )
        {
            if ( other.commandOptions != null )
            {
                return false;
            }
        }
        else if ( !commandOptions.equals( other.commandOptions ) )
        {
            return false;
        }
        if ( enabled == null )
        {
            if ( other.enabled != null )
            {
                return false;
            }
        }
        else if ( !enabled.equals( other.enabled ) )
        {
            return false;
        }
        if ( id == null )
        {
            if ( other.id != null )
            {
                return false;
            }
        }
        else if ( !id.equals( other.id ) )
        {
            return false;
        }
        if ( name == null )
        {
            if ( other.name != null )
            {
                return false;
            }
        }
        else if ( !name.equals( other.name ) )
        {
            return false;
        }
        if ( note == null )
        {
            if ( other.note != null )
            {
                return false;
            }
        }
        else if ( !note.equals( other.note ) )
        {
            return false;
        }
        if ( oneInstance == null )
        {
            if ( other.oneInstance != null )
            {
                return false;
            }
        }
        else if ( !oneInstance.equals( other.oneInstance ) )
        {
            return false;
        }
        if ( sameInstance == null )
        {
            if ( other.sameInstance != null )
            {
                return false;
            }
        }
        else if ( !sameInstance.equals( other.sameInstance ) )
        {
            return false;
        }
        if ( serverAutoSelect == null )
        {
            if ( other.serverAutoSelect != null )
            {
                return false;
            }
        }
        else if ( !serverAutoSelect.equals( other.serverAutoSelect ) )
        {
            return false;
        }
        return true;
    }

}
