/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.i18n;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.MissingResourceException;
import java.util.Objects;
import java.util.ResourceBundle;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.info.TranslationLocale;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.slf4j.Logger;

import com.vaadin.flow.i18n.I18NProvider;

public class TranslationProvider
    implements I18NProvider
{

    private static final long serialVersionUID = 6418059937392578502L;

    private static final Logger LOGGER = LoggerUtil.getLogger();

    public static final TranslationLocale DEFAULT_TRANSLATION_LOCALE = TranslationLocale.DEFAULT_LOCALE;

    private static final String BUNDLE_PREFIX = "org.dussan.vaadin.dvirtual.i18n.translations.translation";

    private static final String ERROR_MISSING_RESOURCE = "Missing resource for %s";

    private static final String ERROR_NULL_VALUE = "Got language request for key with null/empty value!";

    private List<Locale> locales = Collections.unmodifiableList( Arrays.asList( TranslationLocale.EN_US.getLocale(), TranslationLocale.SL_SI.getLocale() ) );

    private TranslationLocale translationLocale = null;

    @Override
    public List<Locale> getProvidedLocales()
    {
        return locales;
    }

    @Override
    public String getTranslation( String key, Locale locale, Object... params )
    {
        if ( StringUtils.isBlank( key ) )
        {
            LOGGER.info( ERROR_NULL_VALUE );
            return StringUtils.EMPTY;
        }

        String value = StringUtils.EMPTY;
        ResourceBundle bundle = null;
        try
        {
            bundle = ResourceBundle.getBundle( BUNDLE_PREFIX, locale );
            value = bundle.getString( key );
        }
        catch ( MissingResourceException | NullPointerException exception )
        {
            LOGGER.error( String.format( ERROR_MISSING_RESOURCE, key ) );
        }

        // get the value from the default locale if it is empty or null
        if ( StringUtils.isBlank( value ) )
        {
            try
            {
                bundle = ResourceBundle.getBundle( BUNDLE_PREFIX, DEFAULT_TRANSLATION_LOCALE.getLocale() );
                value = bundle.getString( key );
            }
            catch ( MissingResourceException | NullPointerException exception )
            {
                LOGGER.error( String.format( ERROR_MISSING_RESOURCE, key ) );
            }
        }

        if ( StringUtils.isNotBlank( value ) && ArrayUtils.isNotEmpty( params ) )
        {
            value = String.format( value, params );
        }
        return StringUtils.defaultIfBlank( value, key );
    }

    public TranslationLocale getTranslationLocale()
    {
        return translationLocale;
    }

    public void setTranslationLocale( TranslationLocale translationLocale )
    {
        this.translationLocale = Objects.requireNonNullElse( translationLocale, DEFAULT_TRANSLATION_LOCALE );
    }

    public String getTranslation( String key, Object... params )
    {
        return getTranslation( key, getTranslationLocale().getLocale(), params );
    }

}
