/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.runners;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketAddress;
import java.nio.file.Path;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.TimeUnit;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.client.Command;
import org.dussan.vaadin.dvirtual.client.Command.XPRA;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.database.data.UserData;
import org.dussan.vaadin.dvirtual.database.data.UserSessionData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.info.ApplicationLocale;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.EncryptionUtil;
import org.dussan.vaadin.dvirtual.utils.LoggerUtil;
import org.dussan.vaadin.dvirtual.utils.NotificationUtil;
import org.dussan.vaadin.dvirtual.utils.SshUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogConnectingUtil;
import org.slf4j.Logger;

import com.jcraft.jsch.Channel;
import com.jcraft.jsch.ChannelExec;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;
import com.vaadin.flow.component.UI;
import com.vaadin.flow.component.notification.NotificationVariant;

public abstract class ApplicationRunner
    implements Runnable
{

    private static final Logger LOGGER = LoggerUtil.getLogger();

    private static final int BYTES_LENGTH = 8192;

    private static final int MAX_WAIT_INTERVAL = 60000; // milliseconds

    private static final int NO_STREAM_VALUE = -1;

    private static final int PING_INTERVAL = 5000; // milliseconds

    private static final int SHOW_DIALOG_SLEEP_INTERVAL = 1000; // milliseconds

    private static final int SLEEP_INTERVAL = 50; // milliseconds

    private static final int SSH_SLEEP_INTERVAL = 10000; // milliseconds

    private static final int STATUS_SLEEP_INTERVAL = 1000; // milliseconds

    private static final String ERROR_APPLICATION = "Application is not installed or wrong command.";

    private static final String ERROR_FREE_PORTS = "No free ports on the Xpra servers.";

    private static final String ERROR_LOCALE = "Locale is not installed.";

    private static final String ERROR_SERVERS = "No available servers.";

    private static final String ERROR_XPRA = "Application Xpra is not installed or wrong command.";

    private static final String ERROR_XPRA_SERVERS = "No available Xpra servers.";

    private static final String FILE_SEPARATOR = File.separator;

    private static final String STRING_MINUS = "-";

    private static final String STRING_REGEX_VALUE = "[^\\d]";

    private static final String STRING_SPACE = " ";

    private static final String TEMP_DIR = System.getProperty( "java.io.tmpdir" );

    private int freeXpraPort = UserSessionData.NEW_PORT;

    private long applicationId = ApplicationData.NEW_APPLICATION;

    private long userSessionId = UserSessionData.NEW_USER_SESSION;

    private String canvasId = UserSessionData.NEW_DATA;

    private List<ServerData> availableServerDatas = null;

    private Database database = null;

    private VirtualUtil virtualUtil = null;

    private NotificationUtil notification20percentUtil = null;

    private NotificationUtil notification5percentUtil = null;

    private NotificationUtil notification60percentUtil = null;

    private NotificationUtil notification90percentUtil = null;

    private NotificationUtil notification100percentUtil = null;

    private NotificationUtil notificationErrorUtil = null;

    private ServerData xpraServerData = null;

    private UI ui = null;

    protected ApplicationRunner( UI ui, Database database, long applicationId, String canvasId, VirtualUtil virtualUtil )
    {
        this.ui = ui;
        this.database = database;
        this.applicationId = applicationId;
        this.canvasId = canvasId;
        this.virtualUtil = virtualUtil;
    }

    public abstract void applicationStarted( long userSessionId, DialogConnectingUtil dialogConnectingUtil );

    private void closeAllNotifications()
    {
        notification5percentUtil.close();
        notification20percentUtil.close();
        notification60percentUtil.close();
        notification90percentUtil.close();
        notification100percentUtil.close();
        notificationErrorUtil.close();
        ui.getPage()
          .executeJs( Command.JS.SHOW_NOTIFICATIONS, false );
    }

    private void stopServer()
    {
        if ( Objects.nonNull( xpraServerData ) && freeXpraPort != UserSessionData.NEW_PORT )
        {
            Channel channel = null;
            Session session = null;
            try
            {
                // start SSH session
                session = SshUtil.createSession( virtualUtil.getUserName(), EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ),
                                                 xpraServerData.getAddress(), xpraServerData.getSshPort() );
                session.connect();

                // stop XPRA server
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( String.format( Command.XPRA.STOP, freeXpraPort ) );
                channel.connect();
                for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( channel.isClosed() )
                    {
                        break;
                    }
                }
                channel.disconnect();

                // stop Xpra pulseaudio
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.KILL_XPRA_PULSEAUDIO, virtualUtil.getUserName() ) );
                channel.connect();
                for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( channel.isClosed() )
                    {
                        break;
                    }
                }
                channel.disconnect();
                session.disconnect();
            }
            catch ( IllegalThreadStateException | InterruptedException | JSchException exception )
            {
                Thread.currentThread()
                      .interrupt();
            }
            finally
            {
                if ( Objects.nonNull( channel ) && channel.isConnected() )
                {
                    channel.disconnect();
                }

                if ( Objects.nonNull( session ) && session.isConnected() )
                {
                    session.disconnect();
                }
            }
        }
    }

    public List<ServerData> getAvailableServerDatas()
    {
        return availableServerDatas;
    }

    public void setAvailableServerDatas( List<ServerData> availableServerDatas )
    {
        this.availableServerDatas = availableServerDatas;
    }

    @Override
    public void run()
    {
        DialogConnectingUtil dialogConnectingUtil = new DialogConnectingUtil( virtualUtil.getTranslation( Translation.APPLICATION_START ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) );
        dialogConnectingUtil.setButtonsVisible( false );
        dialogConnectingUtil.addDetachListener( event -> {
            notification100percentUtil.setDuration( NotificationUtil.NOTIFICATION_CLOSE_DELAY );
            notificationErrorUtil.setDuration( NotificationUtil.NOTIFICATION_CLOSE_LONG_DELAY );
        } );
        ui.getSession()
          .lock();
        ui.access( () -> {
            dialogConnectingUtil.open();
            ui.push();
        } );
        ui.getSession()
          .unlock();

        // notifications
        notification5percentUtil = new NotificationUtil( virtualUtil.getTranslation( Translation.APPLICATION_NOTIFICATION_USER ) );
        notification20percentUtil = new NotificationUtil( virtualUtil.getTranslation( Translation.APPLICATION_NOTIFICATION_INSTANCE_PREPARING ) );
        notification60percentUtil = new NotificationUtil( virtualUtil.getTranslation( Translation.APPLICATION_NOTIFICATION_INSTANCE_STARTING ) );
        notification90percentUtil = new NotificationUtil( virtualUtil.getTranslation( Translation.APPLICATION_NOTIFICATION_STARTING ) );
        notification100percentUtil = new NotificationUtil( virtualUtil.getTranslation( Translation.APPLICATION_NOTIFICATION_RUNNING ) );
        notification100percentUtil.addThemeVariants( NotificationVariant.LUMO_CONTRAST );
        notification100percentUtil.addDetachListener( event -> closeAllNotifications() );
        notificationErrorUtil = new NotificationUtil( virtualUtil.getTranslation( Translation.APPLICATION_NOTIFICATION_ERROR ) );
        notificationErrorUtil.addThemeVariants( NotificationVariant.LUMO_ERROR );
        notificationErrorUtil.addDetachListener( event -> closeAllNotifications() );

        // wait to show dialog before continue
        try
        {
            TimeUnit.MILLISECONDS.sleep( SHOW_DIALOG_SLEEP_INTERVAL );
        }
        catch ( InterruptedException exception )
        {
            Thread.currentThread()
                  .interrupt();
        }

        // get user and application data
        UserData userData = database.getUser( virtualUtil.getUserName() );
        ApplicationData applicationData = database.getApplication( applicationId );
        if ( CollectionUtils.isNotEmpty( getAvailableServerDatas() ) )
        {
            ui.getSession()
              .lock();
            ui.access( () -> {
                dialogConnectingUtil.setProgressValue( DialogConnectingUtil.PROGRESS_5PERCENT );
                if ( virtualUtil.isStartupNotificationsVisible() )
                {
                    notification5percentUtil.open();
                }
                ui.getPage()
                  .executeJs( Command.JS.SHOW_NOTIFICATIONS, true );
                ui.push();
            } );
            ui.getSession()
              .unlock();
        }

        // SSH servers, application, free port
        int serversCounter = 0;
        boolean applicationInstalled = false;
        boolean localeInstalled = false;
        boolean xpraInstalled = false;
        for ( ServerData serverData : getAvailableServerDatas() )
        {
            // progress
            final double serversProgress = ( (double) ++serversCounter / getAvailableServerDatas().size() );
            ui.getSession()
              .lock();
            ui.access( () -> {
                dialogConnectingUtil.setProgressValue( DialogConnectingUtil.PROGRESS_10PERCENT + serversProgress );
                ui.push();
            } );
            ui.getSession()
              .unlock();

            Channel channel = null;
            Session session = null;
            InputStream inputStream = null;
            byte[] tempInputStreamValue = null;
            try (Socket socket = new Socket())
            {
                InetAddress host = InetAddress.getByName( serverData.getAddress() );
                boolean reachableHost = host.isReachable( PING_INTERVAL );
                if ( reachableHost )
                {
                    SocketAddress socketAddress = new InetSocketAddress( serverData.getAddress(), serverData.getSshPort() );
                    socket.connect( socketAddress, PING_INTERVAL );
                    if ( socket.isConnected() )
                    {
                        // start SSH session
                        session = SshUtil.createSession( virtualUtil.getUserName(), EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ),
                                                         serverData.getAddress(), serverData.getSshPort() );
                        session.connect();

                        // find application
                        for ( int timer = 0; timer < SSH_SLEEP_INTERVAL; timer += SLEEP_INTERVAL )
                        {
                            boolean applicationWithPath = StringUtils.contains( applicationData.getCommand(), FILE_SEPARATOR );
                            tempInputStreamValue = new byte[BYTES_LENGTH];
                            channel = session.openChannel( Command.SSH.EXECUTE );
                            ( (ChannelExec) channel ).setCommand( String.format( applicationWithPath ? Command.SSH.FIND_APPLICATION_WITH_PATH : Command.SSH.FIND_APPLICATION,
                                                                                 applicationData.getCommand() ) );
                            inputStream = channel.getInputStream();
                            channel.connect();
                            for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                            {
                                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                                if ( channel.isClosed() )
                                {
                                    break;
                                }
                            }

                            int length = NO_STREAM_VALUE;
                            String inputStreamValue = null;
                            while ( ( length = inputStream.read( tempInputStreamValue, 0, BYTES_LENGTH ) ) > NO_STREAM_VALUE )
                            {
                                if ( length > 0 )
                                {
                                    inputStreamValue = new String( tempInputStreamValue, 0, length );
                                    if ( StringUtils.isNotBlank( inputStreamValue ) )
                                    {
                                        break;
                                    }
                                }
                            }
                            inputStream.close();
                            channel.disconnect();
                            if ( StringUtils.isNotBlank( inputStreamValue ) )
                            {
                                applicationInstalled = true;
                                break;
                            }
                        }

                        // find locale
                        String inputStreamValue = null;
                        for ( int timer = 0; timer < SSH_SLEEP_INTERVAL; timer += SLEEP_INTERVAL )
                        {
                            tempInputStreamValue = new byte[BYTES_LENGTH];
                            channel = session.openChannel( Command.SSH.EXECUTE );
                            ( (ChannelExec) channel ).setCommand( Command.SSH.FIND_LOCALE );
                            inputStream = channel.getInputStream();
                            channel.connect();
                            for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                            {
                                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                                if ( channel.isClosed() )
                                {
                                    break;
                                }
                            }
                            StringBuilder stringBuilder = new StringBuilder();
                            while ( inputStream.available() > 0 )
                            {
                                int length = inputStream.read( tempInputStreamValue, 0, BYTES_LENGTH );
                                if ( length > 0 )
                                {
                                    stringBuilder.append( new String( tempInputStreamValue, 0, length ) );
                                }
                            }
                            inputStreamValue = stringBuilder.toString();
                            inputStream.close();
                            channel.disconnect();
                            if ( StringUtils.isNotBlank( inputStreamValue ) )
                            {
                                break;
                            }
                        }
                        ApplicationLocale applicationLocale = virtualUtil.getApplicationLocale();
                        if ( StringUtils.containsIgnoreCase( inputStreamValue, applicationLocale.getLocale() )
                            || StringUtils.containsIgnoreCase( inputStreamValue, StringUtils.join( applicationLocale.name(), ApplicationLocale.UTF8 ) ) )
                        {
                            localeInstalled = true;
                        }

                        // find Xpra application
                        for ( int timer = 0; timer < SSH_SLEEP_INTERVAL; timer += SLEEP_INTERVAL )
                        {
                            boolean applicationWithPath = StringUtils.contains( Command.XPRA.XPRA_APPLICATION, FILE_SEPARATOR );
                            tempInputStreamValue = new byte[BYTES_LENGTH];
                            channel = session.openChannel( Command.SSH.EXECUTE );
                            ( (ChannelExec) channel ).setCommand( String.format( applicationWithPath ? Command.SSH.FIND_APPLICATION_WITH_PATH : Command.SSH.FIND_APPLICATION,
                                                                                 Command.XPRA.XPRA_APPLICATION ) );
                            inputStream = channel.getInputStream();
                            channel.connect();
                            for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                            {
                                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                                if ( channel.isClosed() )
                                {
                                    break;
                                }
                            }

                            int length = NO_STREAM_VALUE;
                            inputStreamValue = null;
                            while ( ( length = inputStream.read( tempInputStreamValue, 0, BYTES_LENGTH ) ) > NO_STREAM_VALUE )
                            {
                                if ( length > 0 )
                                {
                                    inputStreamValue = new String( tempInputStreamValue, 0, length );
                                    if ( StringUtils.isNotBlank( inputStreamValue ) )
                                    {
                                        break;
                                    }
                                }
                            }
                            inputStream.close();
                            channel.disconnect();
                            if ( StringUtils.isNotBlank( inputStreamValue ) )
                            {
                                xpraInstalled = true;
                                break;
                            }
                        }

                        // free port
                        if ( applicationInstalled && localeInstalled && xpraInstalled && freeXpraPort == UserSessionData.NEW_PORT )
                        {
                            for ( int timer = 0; timer < SSH_SLEEP_INTERVAL; timer += SLEEP_INTERVAL )
                            {
                                tempInputStreamValue = new byte[BYTES_LENGTH];
                                channel = session.openChannel( Command.SSH.EXECUTE );
                                ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.FIND_FREE_XPRA_PORT, serverData.getXpraMinPort(), serverData.getXpraMaxPort() ) );
                                inputStream = channel.getInputStream();
                                channel.connect();
                                for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                                {
                                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                                    if ( channel.isClosed() )
                                    {
                                        break;
                                    }
                                }

                                int length = NO_STREAM_VALUE;
                                while ( ( length = inputStream.read( tempInputStreamValue, 0, BYTES_LENGTH ) ) > NO_STREAM_VALUE )
                                {
                                    if ( length > 0 )
                                    {
                                        String value = new String( tempInputStreamValue, 0, length );
                                        if ( StringUtils.isNotBlank( value ) )
                                        {
                                            String intValue = StringUtils.split( value, virtualUtil.getLineSeparator() )[0];
                                            freeXpraPort = Integer.parseInt( intValue.replaceAll( STRING_REGEX_VALUE, StringUtils.EMPTY ) );
                                            xpraServerData = serverData;
                                            // first remove data from user session data if exists because on the server's free port the application is not running
                                            database.removeUserSession( database.getUserSession( xpraServerData.getId(), freeXpraPort ) );
                                            // and then add new user session data to the database
                                            userSessionId = database.addUserSession( userData, xpraServerData, freeXpraPort, applicationData, null, canvasId, virtualUtil.getDesktopWidth(),
                                                                                     virtualUtil.getDesktopHeight(), UserSessionData.UNDEFINED_WINDOW, UserSessionData.DEFAULT_LOCKED );
                                            break;
                                        }
                                    }
                                }
                                inputStream.close();
                                channel.disconnect();
                                if ( freeXpraPort != UserSessionData.NEW_PORT )
                                {
                                    break;
                                }
                            }
                        }
                        session.disconnect();
                    }
                }
            }
            catch ( InterruptedException | IOException | JSchException exception )
            {
                Thread.currentThread()
                      .interrupt();
            }
            finally
            {
                if ( Objects.nonNull( channel ) && channel.isConnected() )
                {
                    channel.disconnect();
                }

                if ( Objects.nonNull( session ) && session.isConnected() )
                {
                    session.disconnect();
                }
            }

            // getting Xpra server data and free xpra port
            if ( applicationInstalled && localeInstalled && xpraInstalled && Objects.nonNull( xpraServerData ) && freeXpraPort != UserSessionData.NEW_PORT )
            {
                break;
            }
            else
            {
                applicationInstalled = false;
                localeInstalled = false;
                xpraInstalled = false;
            }
        }

        // no available servers
        String errorMessage = null;
        if ( CollectionUtils.isEmpty( getAvailableServerDatas() ) )
        {
            errorMessage = ERROR_SERVERS;
        }

        // locale not installed
        if ( !localeInstalled && StringUtils.isBlank( errorMessage ) )
        {
            errorMessage = ERROR_LOCALE;
        }

        // no Xpra servers
        if ( Objects.isNull( xpraServerData ) && StringUtils.isBlank( errorMessage ) )
        {
            errorMessage = ERROR_XPRA_SERVERS;
        }

        // progress
        ui.getSession()
          .lock();
        ui.access( () -> {
            dialogConnectingUtil.setProgressValue( DialogConnectingUtil.PROGRESS_30PERCENT );
            if ( virtualUtil.isStartupNotificationsVisible() )
            {
                notification20percentUtil.open();
            }
            ui.push();
        } );
        ui.getSession()
          .unlock();

        // application is not installed
        if ( !applicationInstalled && StringUtils.isBlank( errorMessage ) )
        {
            errorMessage = ERROR_APPLICATION;
        }

        // application Xpra is not installed
        if ( !xpraInstalled && StringUtils.isBlank( errorMessage ) )
        {
            errorMessage = ERROR_XPRA;
        }

        // no free port on Xpra servers
        if ( freeXpraPort == UserSessionData.NEW_PORT && StringUtils.isBlank( errorMessage ) )
        {
            errorMessage = ERROR_FREE_PORTS;
        }

        // show message
        if ( StringUtils.isNotBlank( errorMessage ) )
        {
            database.removeUserSession( userSessionId );
            LOGGER.warn( errorMessage );
            ui.getSession()
              .lock();
            ui.access( () -> {
                // for security reasons, a general error message is displayed
                dialogConnectingUtil.setMessage( virtualUtil.getTranslation( Translation.APPLICATION_RUN_ERROR_INSTANCE_FREE ) );
                dialogConnectingUtil.setButtonsVisible( true );
                closeAllNotifications();
                ui.push();
            } );
            ui.getSession()
              .unlock();
            return;
        }

        // progress
        ui.getSession()
          .lock();
        ui.access( () -> {
            dialogConnectingUtil.setProgressValue( DialogConnectingUtil.PROGRESS_50PERCENT );
            ui.push();
        } );
        ui.getSession()
          .unlock();

        // start application on the XPRA server
        Channel channel = null;
        Session session = null;
        boolean applicationStarted = false;
        try
        {
            if ( Objects.isNull( xpraServerData ) )
            {
                throw ( new NullPointerException() );
            }

            // start SSH session
            session = SshUtil.createSession( virtualUtil.getUserName(), EncryptionUtil.decrypt( virtualUtil.getUserPassword(), virtualUtil.getSecretKeySpec(), virtualUtil.getSecureIV() ),
                                             xpraServerData.getAddress(), xpraServerData.getSshPort() );
            session.connect();

            // remove components owned by user if still running
            byte[] temporaryValue = new byte[BYTES_LENGTH];
            channel = session.openChannel( Command.SSH.EXECUTE );
            ( (ChannelExec) channel ).setCommand( Command.XPRA.LIST );
            InputStream inputStream = channel.getInputStream();
            channel.connect();
            for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
            {
                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                if ( channel.isClosed() )
                {
                    break;
                }
            }
            int length = NO_STREAM_VALUE;
            String status = StringUtils.EMPTY;
            while ( ( length = inputStream.read( temporaryValue, 0, BYTES_LENGTH ) ) > NO_STREAM_VALUE )
            {
                if ( length > 0 )
                {
                    status = new String( temporaryValue, 0, length );
                }
            }
            inputStream.close();
            channel.disconnect();
            if ( StringUtils.containsIgnoreCase( status, Command.XPRA.NO_XPRA_SESSION ) )
            {
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.KILL_XPRA_ALL_3MIN, virtualUtil.getUserName() ) );
                channel.connect();
                for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( channel.isClosed() )
                    {
                        break;
                    }
                }
                channel.disconnect();
            }

            // remove zombie processes
            channel = session.openChannel( Command.SSH.EXECUTE );
            ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.KILL_XPRA_ZOMBIE, virtualUtil.getUserName() ) );
            channel.connect();
            for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
            {
                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                if ( channel.isClosed() )
                {
                    break;
                }
            }
            channel.disconnect();

            // create .xpra folder
            String tmpDir = StringUtils.join( Path.of( TEMP_DIR, XPRA.XPRA_APPLICATION ), STRING_MINUS, virtualUtil.getUserName(), STRING_MINUS, freeXpraPort );
            channel = session.openChannel( Command.SSH.EXECUTE );
            ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.CREATE_XPRA_FOLDER, tmpDir ) );
            channel.connect();
            for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
            {
                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                if ( channel.isClosed() )
                {
                    break;
                }
            }
            channel.disconnect();

            // create xpra.conf file
            channel = session.openChannel( Command.SSH.EXECUTE );
            ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.CREATE_XPRA_FILE, tmpDir ) );
            channel.connect();
            for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
            {
                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                if ( channel.isClosed() )
                {
                    break;
                }
            }
            channel.disconnect();

            // init xpra.conf file
            channel = session.openChannel( Command.SSH.EXECUTE );
            ( (ChannelExec) channel ).setCommand( String.format( Command.SSH.INIT_XPRA_FILE, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir, tmpDir,
                                                                 tmpDir, tmpDir, tmpDir ) );
            channel.connect();
            for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
            {
                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                if ( channel.isClosed() )
                {
                    break;
                }
            }
            channel.disconnect();

            // progress
            ui.getSession()
              .lock();
            ui.access( () -> {
                dialogConnectingUtil.setProgressValue( DialogConnectingUtil.PROGRESS_70PERCENT );
                if ( virtualUtil.isStartupNotificationsVisible() )
                {
                    notification60percentUtil.open();
                }
                ui.push();
            } );
            ui.getSession()
              .unlock();

            // start XPRA server
            ApplicationLocale applicationLocale = virtualUtil.getApplicationLocale();
            channel = session.openChannel( Command.SSH.EXECUTE );
            ( (ChannelExec) channel ).setCommand( String.format( Command.XPRA.START, freeXpraPort, xpraServerData.getAddress(), freeXpraPort, applicationLocale.getLocale(), tmpDir,
                                                                 StringUtils.stripToEmpty( StringUtils.join( applicationData.getCommand(), STRING_SPACE, applicationData.getCommandOptions() ) ) ) );
            channel.connect();
            for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += SLEEP_INTERVAL )
            {
                TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                if ( channel.isClosed() )
                {
                    break;
                }
            }
            channel.disconnect();

            // progress
            ui.getSession()
              .lock();
            ui.access( () -> {
                dialogConnectingUtil.setProgressValue( DialogConnectingUtil.PROGRESS_90PERCENT );
                if ( virtualUtil.isStartupNotificationsVisible() )
                {
                    notification90percentUtil.open();
                }
                ui.push();
            } );
            ui.getSession()
              .unlock();

            // get status of started XPRA server
            temporaryValue = new byte[BYTES_LENGTH];
            for ( int timer = 0; timer < MAX_WAIT_INTERVAL; timer += STATUS_SLEEP_INTERVAL )
            {
                channel = session.openChannel( Command.SSH.EXECUTE );
                ( (ChannelExec) channel ).setCommand( Command.XPRA.LIST );
                inputStream = channel.getInputStream();
                channel.connect();
                for ( int t = 0; t < MAX_WAIT_INTERVAL; t += SLEEP_INTERVAL )
                {
                    TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                    if ( channel.isClosed() )
                    {
                        break;
                    }
                }

                length = NO_STREAM_VALUE;
                while ( ( length = inputStream.read( temporaryValue, 0, BYTES_LENGTH ) ) > NO_STREAM_VALUE )
                {
                    if ( length > 0 )
                    {
                        status = new String( temporaryValue, 0, length );
                        if ( StringUtils.contains( status, String.format( Command.XPRA.LIST_STATUS, freeXpraPort ) ) )
                        {
                            applicationStarted = true;
                            break;
                        }
                    }
                }
                inputStream.close();
                channel.disconnect();

                if ( applicationStarted )
                {
                    // update XPRA version
                    for ( int t = 0; t < SSH_SLEEP_INTERVAL; t += SLEEP_INTERVAL )
                    {
                        byte[] tempInputStreamValue = new byte[BYTES_LENGTH];
                        channel = session.openChannel( Command.SSH.EXECUTE );
                        ( (ChannelExec) channel ).setCommand( String.format( Command.XPRA.VERSION, freeXpraPort ) );
                        OutputStream outputStream = channel.getOutputStream();
                        inputStream = channel.getInputStream();
                        channel.connect();
                        for ( int tt = 0; tt < MAX_WAIT_INTERVAL; tt += SLEEP_INTERVAL )
                        {
                            TimeUnit.MILLISECONDS.sleep( SLEEP_INTERVAL );
                            if ( channel.isClosed() )
                            {
                                break;
                            }
                        }

                        length = NO_STREAM_VALUE;
                        String xpraVersion = StringUtils.EMPTY;
                        while ( ( length = inputStream.read( tempInputStreamValue, 0, BYTES_LENGTH ) ) > NO_STREAM_VALUE )
                        {
                            if ( length > 0 )
                            {
                                String version = new String( tempInputStreamValue, 0, length );
                                if ( StringUtils.isNotBlank( version ) && version.length() >= 5 && StringUtils.isNumeric( version.replaceAll( STRING_REGEX_VALUE, StringUtils.EMPTY ) ) )
                                {
                                    xpraVersion = version;
                                    break;
                                }
                            }
                        }
                        outputStream.close();
                        inputStream.close();
                        channel.disconnect();

                        if ( StringUtils.isNotBlank( xpraVersion ) )
                        {
                            xpraServerData.setXpraVersion( xpraVersion );
                            database.updateServer( xpraServerData );
                            break;
                        }
                    }

                    ui.getSession()
                      .lock();
                    ui.access( () -> {
                        applicationStarted( userSessionId, dialogConnectingUtil );
                        dialogConnectingUtil.setProgressValue( DialogConnectingUtil.PROGRESS_STOP );
                        if ( virtualUtil.isStartupNotificationsVisible() )
                        {
                            notification100percentUtil.open();
                        }
                        ui.push();
                    } );
                    ui.getSession()
                      .unlock();
                    break;
                }
            }
            session.disconnect();
        }
        catch ( IllegalThreadStateException | InterruptedException | IOException | NullPointerException | JSchException exception )
        {
            stopServer();
            Thread.currentThread()
                  .interrupt();
        }
        finally
        {
            if ( Objects.nonNull( channel ) && channel.isConnected() )
            {
                channel.disconnect();
            }

            if ( Objects.nonNull( session ) && session.isConnected() )
            {
                session.disconnect();
            }
        }

        // application not started
        if ( !applicationStarted )
        {
            database.removeUserSession( userSessionId );
            ui.getSession()
              .lock();
            ui.access( () -> {
                dialogConnectingUtil.setMessage( virtualUtil.getTranslation( Translation.APPLICATION_RUN_ERROR_START ) );
                dialogConnectingUtil.setButtonsVisible( true );
                if ( virtualUtil.isStartupNotificationsVisible() )
                {
                    notificationErrorUtil.open();
                }
                ui.push();
            } );
            ui.getSession()
              .unlock();
        }
    }

}
