/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.ui.bars.applications;

import java.io.ByteArrayInputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.client.Command;
import org.dussan.vaadin.dvirtual.database.Database;
import org.dussan.vaadin.dvirtual.database.data.ApplicationData;
import org.dussan.vaadin.dvirtual.database.data.ApplicationsGroupData;
import org.dussan.vaadin.dvirtual.database.data.IconData;
import org.dussan.vaadin.dvirtual.database.data.ServerData;
import org.dussan.vaadin.dvirtual.database.data.ServersGroupData;
import org.dussan.vaadin.dvirtual.database.data.UserApplicationData;
import org.dussan.vaadin.dvirtual.database.data.UserApplicationsGroupData;
import org.dussan.vaadin.dvirtual.database.data.UserData;
import org.dussan.vaadin.dvirtual.database.data.UserSessionApplicationData;
import org.dussan.vaadin.dvirtual.database.data.UserSessionData;
import org.dussan.vaadin.dvirtual.i18n.Translation;
import org.dussan.vaadin.dvirtual.runners.ApplicationRunner;
import org.dussan.vaadin.dvirtual.runners.TooltipRunner;
import org.dussan.vaadin.dvirtual.utils.VirtualUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogConnectingUtil;
import org.dussan.vaadin.dvirtual.utils.dialog.DialogInfoUtil;

import com.vaadin.flow.component.DetachEvent;
import com.vaadin.flow.component.UI;
import com.vaadin.flow.component.html.Div;
import com.vaadin.flow.component.html.Image;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.icon.IronIcon;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.tabs.Tab;
import com.vaadin.flow.component.tabs.Tabs;
import com.vaadin.flow.server.StreamResource;

public abstract class ApplicationsBar
    extends HorizontalLayout
{

    private static final long serialVersionUID = 5943799470555438114L;

    private static final int FIRST_ELEMENT = 0;

    private static final int NO_APPLICATIONS = 0;

    private static final int NO_SERVERS = 0;

    private static final int ONLY_ONE_SERVER = 1;

    private static final Integer UNSELECTED_TAB = -1;

    private static final Integer UNSELECTED_WINDOW = -1;

    private static final String ATTRIBUTE_THEME = "theme";

    private static final String ATTRIBUTE_THEME_VALUE = "dark";

    private static final String CLASS_APPLICATION = "application";

    private static final String CLASS_APPLICATIONS_BAR = "applications-bar";

    private static final String CLASS_ICON = "icon";

    private static final String CLASS_LABEL = "label";

    private static final String ICON_NAME_SEPARATOR = ":";

    private static final String MAXIMIZED_NO_INDICATOR = "-%s;";

    private static final String MAXIMIZED_YES_INDICATOR = "+%s;";

    private static final String STRING_ARRAY_SPLITTER = ";";

    private boolean refreshingBar = false;

    private String canvasId = null;

    private transient ApplicationRunner applicationRunner = null;

    private Database database = null;

    private UserData userData = null;

    private VirtualUtil virtualUtil = null;

    private Tabs tabs = null;

    protected ApplicationsBar( Database database, String canvasId, VirtualUtil virtualUtil )
    {
        setJustifyContentMode( JustifyContentMode.CENTER );
        setMargin( false );
        setPadding( false );
        setSpacing( false );
        setSizeUndefined();
        addClassNames( CLASS_APPLICATIONS_BAR );
        this.database = database;
        this.canvasId = canvasId;
        this.virtualUtil = virtualUtil;

        // application bar
        tabs = new Tabs();
        tabs.getElement()
            .setAttribute( ATTRIBUTE_THEME, ATTRIBUTE_THEME_VALUE );
        add( tabs );
    }

    public abstract void runApplication( long userSessionId, DialogConnectingUtil dialogConnectingUtil );

    public abstract void closeChildWindows();

    public abstract void detachOldApplication();

    private List<ServerData> getAvailableServers( ApplicationData applicationData, List<UserSessionData> userSessionDatas, boolean onlyFreeServers )
    {
        // get all enabled servers from application
        List<Long> availableServerIds = new ArrayList<>();
        for ( ServersGroupData serversGroupData : applicationData.getServersGroupDatas() )
        {
            availableServerIds.addAll( serversGroupData.getServerDatas()
                                                       .stream()
                                                       .filter( Objects::nonNull )
                                                       .filter( ServerData::isEnabled )
                                                       .mapToLong( ServerData::getId )
                                                       .boxed()
                                                       .collect( Collectors.toList() ) );
        }

        // free servers: counter of current users <= maximum users
        List<Long> serverWithMaxUsersIds = CollectionUtils.isEmpty( availableServerIds ) ? new ArrayList<>() : database.getServerWithMaxUsersIds( availableServerIds.toArray( Long[]::new ) );
        List<Long> freeServerIds = new ArrayList<>( CollectionUtils.removeAll( availableServerIds, serverWithMaxUsersIds ) );

        if ( applicationData.isServerAutoSelect() && applicationData.isNotSameInstance() )
        {
            List<Long> ids = new ArrayList<>( freeServerIds );
            userSessionDatas.forEach( userSessionData -> freeServerIds.removeIf( freeServerId -> Objects.equals( userSessionData.getApplicationData(), applicationData )
                && Objects.equals( freeServerId, userSessionData.getServerData()
                                                                .getId() ) ) );
            if ( CollectionUtils.isNotEmpty( freeServerIds ) )
            {
                return database.getServers( freeServerIds.toArray( Long[]::new ) );
            }
            else if ( CollectionUtils.isNotEmpty( ids ) )
            {
                return database.getServers( ids.toArray( Long[]::new ) );
            }
        }
        else if ( CollectionUtils.isNotEmpty( freeServerIds ) && onlyFreeServers )
        {
            userSessionDatas.forEach( userSessionData -> freeServerIds.removeIf( freeServerId -> Objects.equals( userSessionData.getApplicationData(), applicationData )
                && Objects.equals( freeServerId, userSessionData.getServerData()
                                                                .getId() ) ) );
            if ( CollectionUtils.isNotEmpty( freeServerIds ) )
            {
                return database.getServers( freeServerIds.toArray( Long[]::new ) );
            }
        }
        else if ( CollectionUtils.isNotEmpty( freeServerIds ) )
        {
            return database.getServers( freeServerIds.toArray( Long[]::new ) );
        }

        // no available servers
        return ( new ArrayList<>() );
    }

    private void addApplications()
    {
        List<ApplicationData> applicationDatas = new ArrayList<>();
        applicationDatas.addAll( database.getUserApplications( userData )
                                         .stream()
                                         .filter( Objects::nonNull )
                                         .map( UserApplicationData::getApplicationData )
                                         .filter( ApplicationData::isEnabled )
                                         .collect( Collectors.toList() ) );
        applicationDatas.addAll( database.getUserApplicationGroups( userData )
                                         .stream()
                                         .filter( Objects::nonNull )
                                         .map( UserApplicationsGroupData::getApplicationsGroupData )
                                         .filter( ApplicationsGroupData::isEnabled )
                                         .flatMap( data -> data.getApplicationDatas()
                                                               .stream() )
                                         .collect( Collectors.toList() ) );
        // add applications to bar
        applicationDatas.stream()
                        .distinct()
                        .sorted( Comparator.comparing( ApplicationData::getName ) )
                        .forEach( applicationData -> {
                            Div div = new Div();

                            // icon
                            IconData iconData = applicationData.getIconData();
                            if ( Objects.nonNull( iconData ) )
                            {
                                if ( Objects.isNull( iconData.getIcon() ) )
                                {
                                    String iconName = iconData.getName();
                                    IronIcon ironIcon = new IronIcon( iconName.contains( ICON_NAME_SEPARATOR ) ? iconName.split( ICON_NAME_SEPARATOR )[0] : iconName,
                                                                      iconName.contains( ICON_NAME_SEPARATOR ) ? iconName.split( ICON_NAME_SEPARATOR )[1] : iconName );
                                    ironIcon.addClassName( CLASS_ICON );
                                    div.add( ironIcon );
                                }
                                else
                                {
                                    StreamResource streamResource = new StreamResource( iconData.getName(), () -> new ByteArrayInputStream( iconData.getIcon() ) );
                                    Image image = new Image( streamResource, iconData.getName() );
                                    image.addClassName( CLASS_ICON );
                                    div.add( image );
                                }
                            }

                            // label
                            Label label = new Label( applicationData.getName() );
                            label.addClassName( CLASS_LABEL );
                            div.add( label );

                            // tab
                            Tab tab = new Tab( div );
                            tab.addClassName( CLASS_APPLICATION );
                            tab.setId( Long.toString( applicationData.getId() ) );
                            tabs.add( tab );

                            if ( virtualUtil.isTooltipsVisible() )
                            {
                                TooltipRunner tooltipRunner = new TooltipRunner( tab, StringUtils.defaultIfBlank( applicationData.getNote(), applicationData.getName() ), UI.getCurrent() );
                                virtualUtil.startRunnerService( tooltipRunner );
                            }
                        } );

        // run application
        tabs.setSelectedIndex( UNSELECTED_TAB );
        tabs.addSelectedChangeListener( event -> {
            Tab tab = event.getSource()
                           .getSelectedTab();
            closeChildWindows();
            if ( !refreshingBar && Objects.nonNull( tab ) )
            {
                String applicationIdString = tab.getId()
                                                .orElse( Integer.toString( UNSELECTED_TAB ) );
                if ( StringUtils.isNumeric( applicationIdString ) )
                {
                    int applicationId = Integer.parseInt( applicationIdString );
                    if ( applicationId != UNSELECTED_TAB )
                    {
                        selectApplication( userData, applicationId );
                    }
                    tabs.setSelectedIndex( UNSELECTED_TAB );
                }
            }

            // enable applications bar
            UI.getCurrent()
              .getPage()
              .executeJs( Command.JS.DISABLE_BARS, false );
        } );
    }

    private void removeApplications()
    {
        tabs.removeAll();
    }

    public void setUserData( UserData userData )
    {
        this.userData = userData;
        addApplications();
    }

    public void selectApplication( UserData userData, long applicationId )
    {
        // check maximum allowed running applications
        List<UserSessionData> userSessionDatas = database.getAllUserSessions( canvasId );
        int numberOfRunningApplications = NO_APPLICATIONS;
        for ( UserSessionData userSessionData : userSessionDatas )
        {
            numberOfRunningApplications += StringUtils.countMatches( userSessionData.getStatus(), STRING_ARRAY_SPLITTER );
        }

        ApplicationData applicationData = database.getApplication( applicationId );
        userSessionDatas = database.getAllUserSessions( userData );
        if ( Objects.nonNull( userData ) && Objects.nonNull( applicationData ) )
        {
            List<UserSessionApplicationData> userSessionApplicationDatas = database.getUserSessionApplications( userData );

            // all application running in same instance
            boolean sameInstanceActive = false;
            long sameInstanceUserSessionId = UserSessionData.NEW_USER_SESSION;
            if ( virtualUtil.isAllRunningInSameInstance() && CollectionUtils.isNotEmpty( userSessionDatas ) )
            {
                applicationData.setOneInstance( applicationData.isSameInstance() || applicationData.isOneInstance() ? ApplicationData.ONE_INSTANCE_ENABLED : applicationData.getOneInstance() );
                UserSessionData userSessionData = userSessionDatas.get( FIRST_ELEMENT );
                sameInstanceUserSessionId = StringUtils.equals( canvasId, userSessionData.getCanvasId() ) ? userSessionData.getId() : sameInstanceUserSessionId;
                sameInstanceActive = true;
            }

            // check application's same instance/check free servers
            boolean applicationSameInstanceActive = false;
            if ( !sameInstanceActive && CollectionUtils.isNotEmpty( userSessionDatas ) )
            {
                if ( numberOfRunningApplications >= userData.getMaxApplications() )
                {
                    applicationData.setOneInstance( ApplicationData.ONE_INSTANCE_ENABLED );
                }
                else
                {
                    final List<UserSessionData> sessionDatas = userSessionDatas;
                    final List<ServerData> availableServers = getAvailableServers( applicationData, userSessionDatas, false );
                    final List<ServerData> allActiveServers = userSessionDatas.stream()
                                                                              .filter( data -> Objects.nonNull( data.getWindowId() ) )
                                                                              .filter( data -> data.getApplicationData()
                                                                                                   .getId() == applicationId )
                                                                              .map( UserSessionData::getServerData )
                                                                              .collect( Collectors.toList() );
                    userSessionApplicationDatas.stream()
                                               .filter( data -> data.getApplicationId() == applicationId )
                                               .forEach( data -> sessionDatas.stream()
                                                                             .filter( d -> Objects.equals( d.getId(), data.getUserSessionId() ) )
                                                                             .forEach( d -> allActiveServers.add( d.getServerData() ) ) );
                    List<ServerData> activeServers = new ArrayList<>( new LinkedHashSet<>( allActiveServers ) );
                    if ( applicationData.isSameInstance()
                                        .booleanValue() )
                    {
                        // check application's same instance
                        applicationSameInstanceActive = availableServers.size() > NO_SERVERS && activeServers.size() == availableServers.size();
                    }
                    else if ( applicationData.isNotSameInstance()
                                             .booleanValue()
                        && availableServers.size() > NO_SERVERS && activeServers.size() == availableServers.size() && CollectionUtils.containsAll( activeServers, availableServers ) )
                    {
                        // check free servers
                        UserSessionData userSessionData = userSessionDatas.stream()
                                                                          .filter( data -> CollectionUtils.containsAny( activeServers, data.getServerData() ) )
                                                                          .reduce( ( first, second ) -> second )
                                                                          .orElse( null );
                        if ( Objects.nonNull( userSessionData ) )
                        {
                            applicationData.setOneInstance( applicationData.isSameInstance() || applicationData.isOneInstance() ? ApplicationData.ONE_INSTANCE_ENABLED
                                            : applicationData.getOneInstance() );
                            sameInstanceUserSessionId = StringUtils.equals( canvasId, userSessionData.getCanvasId() ) ? userSessionData.getId() : sameInstanceUserSessionId;
                            sameInstanceActive = true;
                        }
                    }
                }
            }

            // check one instance
            boolean oneInstanceActive = false;
            long oneInstanceUserSessionId = UserSessionData.NEW_USER_SESSION;
            if ( CollectionUtils.isNotEmpty( userSessionDatas ) && ( applicationSameInstanceActive || applicationData.isOneInstance() ) )
            {
                int newWindowFocusId = UNSELECTED_WINDOW;
                List<UserSessionData> sessionDatas = userSessionDatas.stream()
                                                                     .sorted( Comparator.comparing( UserSessionData::getUpdated )
                                                                                        .reversed() )
                                                                     .collect( Collectors.toList() );
                for ( UserSessionData userSessionData : sessionDatas )
                {
                    // apllication's focused window id
                    int windowId = UNSELECTED_WINDOW;
                    String windowsStatus = StringUtils.defaultIfBlank( userSessionData.getStatus(), StringUtils.EMPTY );
                    if ( userSessionData.getApplicationData()
                                        .getId() != applicationId )
                    {
                        windowsStatus = StringUtils.remove( windowsStatus, String.format( MAXIMIZED_NO_INDICATOR, userSessionData.getWindowId() ) );
                        windowsStatus = StringUtils.remove( windowsStatus, String.format( MAXIMIZED_YES_INDICATOR, userSessionData.getWindowId() ) );
                    }
                    for ( UserSessionApplicationData userSessionApplicationData : userSessionApplicationDatas )
                    {
                        if ( Objects.equals( userSessionApplicationData.getUserSessionId(), userSessionData.getId() ) && userSessionApplicationData.getApplicationId() != applicationId )
                        {
                            windowsStatus = StringUtils.remove( windowsStatus, String.format( MAXIMIZED_NO_INDICATOR, userSessionApplicationData.getWindowId() ) );
                            windowsStatus = StringUtils.remove( windowsStatus, String.format( MAXIMIZED_YES_INDICATOR, userSessionApplicationData.getWindowId() ) );
                        }
                    }
                    if ( StringUtils.isNotBlank( windowsStatus ) )
                    {
                        String[] windowIds = StringUtils.split( windowsStatus, STRING_ARRAY_SPLITTER );
                        windowId = Math.abs( Integer.parseInt( windowIds[windowIds.length - 1] ) );
                    }

                    // one instance
                    if ( Objects.nonNull( userSessionData.getWindowId() ) && userSessionData.getWindowId() == windowId )
                    {
                        oneInstanceUserSessionId = StringUtils.equals( canvasId, userSessionData.getCanvasId() ) ? userSessionData.getId() : oneInstanceUserSessionId;
                        newWindowFocusId = windowId;
                        oneInstanceActive = true;
                    }
                    else
                    {
                        final int winId = windowId;
                        final long userSessionId = userSessionData.getId();
                        UserSessionApplicationData userSessionApplicationData = userSessionApplicationDatas.stream()
                                                                                                           .filter( data -> data.getUserSessionId() == userSessionId )
                                                                                                           .filter( data -> data.getApplicationId() == applicationId )
                                                                                                           .filter( data -> data.getWindowId() == winId )
                                                                                                           .findFirst()
                                                                                                           .orElse( null );
                        if ( Objects.nonNull( userSessionApplicationData ) )
                        {
                            UserSessionData sessionData = userSessionDatas.stream()
                                                                          .filter( data -> Objects.equals( data.getId(), userSessionApplicationData.getUserSessionId() ) )
                                                                          .findFirst()
                                                                          .orElse( null );
                            oneInstanceUserSessionId = Objects.isNull( sessionData ) ? null : oneInstanceUserSessionId;
                            if ( Objects.nonNull( sessionData ) )
                            {
                                oneInstanceUserSessionId = StringUtils.equals( canvasId, sessionData.getCanvasId() ) ? sessionData.getId() : oneInstanceUserSessionId;
                                newWindowFocusId = StringUtils.equals( canvasId, sessionData.getCanvasId() ) ? windowId : newWindowFocusId;
                            }
                            oneInstanceActive = true;
                        }
                    }

                    if ( oneInstanceActive && newWindowFocusId != UNSELECTED_WINDOW )
                    {
                        // update windows focus: bring active window to top
                        String noMaximized = String.format( MAXIMIZED_NO_INDICATOR, newWindowFocusId );
                        String yesMaximized = String.format( MAXIMIZED_YES_INDICATOR, newWindowFocusId );
                        String status = StringUtils.defaultIfBlank( userSessionData.getStatus(), StringUtils.EMPTY );
                        boolean maximized = StringUtils.contains( status, yesMaximized );
                        status = StringUtils.remove( status, noMaximized );
                        status = StringUtils.remove( status, yesMaximized );
                        status = StringUtils.join( status, maximized ? yesMaximized : noMaximized );
                        userSessionData.setStatus( status );
                        userSessionData.setUpdated( new Date() );
                        database.updateUserSession( userSessionData );
                    }
                    if ( oneInstanceActive )
                    {
                        break;
                    }
                }
            }

            // check same instance in applications group
            if ( !oneInstanceActive && !sameInstanceActive && CollectionUtils.isNotEmpty( userSessionDatas ) )
            {
                // applications group without same instance
                if ( database.isApplicationGroupsWithSameInstanceEmpty() && ( applicationData.isSameInstance() || applicationData.isOneInstance() ) )
                {
                    UserSessionData userSessionData = userSessionDatas.stream()
                                                                      .filter( data -> data.getApplicationData()
                                                                                           .getId() != applicationId )
                                                                      .filter( data -> data.getApplicationData()
                                                                                           .isSameInstance()
                                                                          || data.getApplicationData()
                                                                                 .isOneInstance() )
                                                                      .findFirst()
                                                                      .orElse( null );
                    if ( Objects.nonNull( userSessionData ) )
                    {
                        sameInstanceUserSessionId = StringUtils.equals( canvasId, userSessionData.getCanvasId() ) ? userSessionData.getId() : sameInstanceUserSessionId;
                        sameInstanceActive = true;
                    }
                    else
                    {
                        Map<Long, ApplicationData> applicationDatas = new HashMap<>();
                        MapUtils.populateMap( applicationDatas, database.getAllApplications(), ApplicationData::getId );
                        UserSessionApplicationData userSessionApplicationData = userSessionApplicationDatas.stream()
                                                                                                           .filter( data -> data.getApplicationId() != applicationId )
                                                                                                           .filter( data -> applicationDatas.get( data.getApplicationId() )
                                                                                                                                            .isSameInstance()
                                                                                                               || applicationDatas.get( data.getApplicationId() )
                                                                                                                                  .isOneInstance() )
                                                                                                           .findFirst()
                                                                                                           .orElse( null );
                        if ( Objects.nonNull( userSessionApplicationData ) )
                        {
                            userSessionData = userSessionDatas.stream()
                                                              .filter( data -> Objects.equals( data.getId(), userSessionApplicationData.getUserSessionId() ) )
                                                              .findFirst()
                                                              .orElse( null );
                            sameInstanceUserSessionId = Objects.isNull( userSessionData ) ? null : sameInstanceUserSessionId;
                            if ( Objects.nonNull( userSessionData ) )
                            {
                                sameInstanceUserSessionId = StringUtils.equals( canvasId, userSessionData.getCanvasId() ) ? userSessionData.getId() : sameInstanceUserSessionId;
                            }
                            sameInstanceActive = true;
                        }
                    }
                }

                // applications group with same instance
                if ( !sameInstanceActive )
                {
                    Set<Long> applicationIds = new HashSet<>();
                    database.getApplicationGroupsWithSameInstance( applicationData )
                            .forEach( data -> data.getApplicationDatas()
                                                  .forEach( d -> applicationIds.add( d.getId() ) ) );
                    UserSessionData userSessionData = userSessionDatas.stream()
                                                                      .filter( data -> StringUtils.equals( canvasId, data.getCanvasId() ) )
                                                                      .filter( data -> CollectionUtils.containsAny( applicationIds, data.getApplicationData()
                                                                                                                                        .getId() ) )
                                                                      .sorted( Comparator.comparing( UserSessionData::getUpdated )
                                                                                         .reversed() )
                                                                      .findFirst()
                                                                      .orElse( null );
                    if ( Objects.nonNull( userSessionData ) )
                    {
                        UserSessionData sessionData = userSessionDatas.stream()
                                                                      .filter( data -> Objects.nonNull( data.getWindowId() ) )
                                                                      .filter( data -> data.getApplicationData()
                                                                                           .getId() == applicationId )
                                                                      .findFirst()
                                                                      .orElse( null );
                        if ( Objects.nonNull( sessionData ) )
                        {
                            if ( applicationData.isSameInstance()
                                                .booleanValue() )
                            {
                                oneInstanceUserSessionId = StringUtils.equals( canvasId, sessionData.getCanvasId() ) ? sessionData.getId() : oneInstanceUserSessionId;
                                oneInstanceActive = true;
                            }
                            else
                            {
                                sameInstanceUserSessionId = StringUtils.equals( canvasId, sessionData.getCanvasId() ) ? sessionData.getId() : sameInstanceUserSessionId;
                                sameInstanceActive = true;
                            }
                        }
                        else
                        {
                            UserSessionApplicationData userSessionApplicationData = userSessionApplicationDatas.stream()
                                                                                                               .filter( data -> data.getApplicationId() == applicationId )
                                                                                                               .findFirst()
                                                                                                               .orElse( null );
                            if ( Objects.nonNull( userSessionApplicationData ) )
                            {
                                sessionData = userSessionDatas.stream()
                                                              .filter( data -> Objects.equals( data.getId(), userSessionApplicationData.getUserSessionId() ) )
                                                              .findFirst()
                                                              .orElse( null );
                                if ( applicationData.isSameInstance()
                                                    .booleanValue() )
                                {
                                    oneInstanceUserSessionId = Objects.isNull( sessionData ) ? null : oneInstanceUserSessionId;
                                    if ( Objects.nonNull( sessionData ) )
                                    {
                                        oneInstanceUserSessionId = StringUtils.equals( canvasId, sessionData.getCanvasId() ) ? sessionData.getId() : oneInstanceUserSessionId;
                                    }
                                    oneInstanceActive = true;
                                }
                                else
                                {
                                    sameInstanceUserSessionId = Objects.isNull( sessionData ) ? null : sameInstanceUserSessionId;
                                    if ( Objects.nonNull( sessionData ) )
                                    {
                                        sameInstanceUserSessionId = StringUtils.equals( canvasId, sessionData.getCanvasId() ) ? sessionData.getId() : oneInstanceUserSessionId;
                                    }
                                    sameInstanceActive = true;
                                }
                            }
                        }
                        if ( !oneInstanceActive && !sameInstanceActive )
                        {
                            sameInstanceUserSessionId = userSessionData.getId();
                            sameInstanceActive = true;
                        }
                    }
                    else
                    {
                        userSessionData = userSessionDatas.stream()
                                                          .filter( data -> !StringUtils.equals( canvasId, data.getCanvasId() ) )
                                                          .filter( data -> CollectionUtils.containsAny( applicationIds, data.getApplicationData()
                                                                                                                            .getId() ) )
                                                          .findFirst()
                                                          .orElse( null );
                        if ( Objects.nonNull( userSessionData ) )
                        {
                            sameInstanceActive = true;
                        }
                    }
                }
            }

            virtualUtil.setRecreateNewWindowEnabled( false );
            if ( oneInstanceActive )
            {
                if ( Objects.isNull( oneInstanceUserSessionId ) )
                {
                    new DialogInfoUtil( virtualUtil.getTranslation( Translation.APPLICATION_ERROR_CONNECT ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                }
                else if ( oneInstanceUserSessionId == UserSessionData.NEW_USER_SESSION )
                {
                    new DialogInfoUtil( virtualUtil.getTranslation( Translation.APPLICATION_ERROR_INSTANCE_ONE, applicationData.getName() ),
                                        virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                }
                else
                {
                    runApplication( oneInstanceUserSessionId, null );
                }
            }
            else if ( numberOfRunningApplications >= userData.getMaxApplications() )
            {
                new DialogInfoUtil( virtualUtil.getTranslation( Translation.DATA_ERROR_ALLOWED ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
            }
            else if ( sameInstanceActive )
            {
                if ( Objects.isNull( sameInstanceUserSessionId ) )
                {
                    new DialogInfoUtil( virtualUtil.getTranslation( Translation.APPLICATION_ERROR_CONNECT ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                }
                else if ( sameInstanceUserSessionId == UserSessionData.NEW_USER_SESSION )
                {
                    new DialogInfoUtil( virtualUtil.isAllRunningInSameInstance() ? virtualUtil.getTranslation( Translation.APPLICATION_ERROR_INSTANCE_FREE )
                                    : virtualUtil.getTranslation( Translation.APPLICATION_ERROR_INSTANCE_ONE, applicationData.getName() ),
                                        virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                }
                else
                {
                    virtualUtil.setChildApplicationId( applicationId );
                    runApplication( sameInstanceUserSessionId, null );
                }
            }
            else
            {
                List<ServerData> serverDatas = getAvailableServers( applicationData, userSessionDatas, true );
                if ( serverDatas.size() == NO_SERVERS )
                {
                    new DialogInfoUtil( virtualUtil.getTranslation( Translation.APPLICATION_ERROR_INSTANCE_FREE ), virtualUtil.getTranslation( Translation.BUTTON_CLOSE ) ).open();
                }
                else
                {
                    if ( applicationData.isServerAutoSelect()
                                        .booleanValue()
                        || serverDatas.size() == ONLY_ONE_SERVER )
                    {
                        Collections.shuffle( serverDatas );
                        applicationRunner = new ApplicationRunner( UI.getCurrent(), database, applicationId, canvasId, virtualUtil )
                        {
                            @Override
                            public void applicationStarted( long userSessionId, DialogConnectingUtil dialogConnectingUtil )
                            {
                                detachOldApplication();
                                runApplication( userSessionId, dialogConnectingUtil );
                            }
                        };
                        applicationRunner.setAvailableServerDatas( serverDatas );
                        virtualUtil.startRunnerService( applicationRunner );
                    }
                    else
                    {
                        Collections.sort( serverDatas, ( ServerData data1, ServerData data2 ) -> data1.getName()
                                                                                                      .compareTo( data2.getName() ) );
                        SelectServerDialog selectServerDialog = new SelectServerDialog( serverDatas, virtualUtil )
                        {

                            private static final long serialVersionUID = -8182638865697521049L;

                            @Override
                            public void setSelectedServers( List<ServerData> selectedServerDatas )
                            {
                                applicationRunner = new ApplicationRunner( UI.getCurrent(), database, applicationId, canvasId, virtualUtil )
                                {
                                    @Override
                                    public void applicationStarted( long userSessionId, DialogConnectingUtil dialogConnectingUtil )
                                    {
                                        detachOldApplication();
                                        runApplication( userSessionId, dialogConnectingUtil );
                                    }
                                };
                                applicationRunner.setAvailableServerDatas( selectedServerDatas );
                                virtualUtil.startRunnerService( applicationRunner );
                            }
                        };
                        selectServerDialog.open();
                    }
                }
            }
        }
    }

    public void refresh()
    {
        refreshingBar = true;
        removeApplications();
        addApplications();
        refreshingBar = false;
    }

    public boolean isEmpty()
    {
        return ( tabs.getComponentCount() == NO_APPLICATIONS );
    }

    public boolean isNotEmpty()
    {
        return !isEmpty();
    }

    @Override
    protected void onDetach( DetachEvent detachEvent )
    {
        removeApplications();
    }

}
