/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dvirtual.xpra.io;

import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.Arrays;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class HeaderChunk
    implements Serializable
{

    private static final long serialVersionUID = -5161640570666125686L;

    private static final Logger LOGGER = LoggerFactory.getLogger( HeaderChunk.class );

    public static final int FLAGS_BENCODE = 0x0;

    public static final int FLAGS_CIPHER = 0x2;

    public static final int FLAGS_NOHEADER = 0x10000;

    public static final int FLAGS_RENCODE = 0x1;

    public static final int FLAGS_YAML = 0x4;

    public static final int BROTLI_FLAG = 0x40;

    public static final int LZ4_FLAG = 0x10;

    public static final int LZO_FLAG = 0x20;

    public static final int MAX_PACKET_INDEX = 20;

    public static final int ZLIB_FLAG = 0x0;

    private static final String FAILED_TO_READ_HEADER = "Failed to read header.";

    private static final int HEADER_SIZE = 8;

    private static final byte MAGIC_BYTE = 'P';

    private boolean badHeader = false;

    private byte[] header = null;

    public HeaderChunk()
    {
        header = new byte[HEADER_SIZE];
        header[0] = MAGIC_BYTE;
    }

    public void readHeader( InputStream inputStream )
        throws IOException
    {
        int headerRead = 0;
        while ( headerRead < HEADER_SIZE )
        {
            final int bytesRead = inputStream.read( header, headerRead, HEADER_SIZE - headerRead );
            if ( bytesRead < 0 )
            {
                LOGGER.error( FAILED_TO_READ_HEADER );
                throw new EOFException( FAILED_TO_READ_HEADER );
            }
            headerRead += bytesRead;
        }

        if ( header[0] != MAGIC_BYTE || getPacketIndex() >= MAX_PACKET_INDEX || hasFlags( FLAGS_CIPHER | FLAGS_YAML | FLAGS_NOHEADER ) || isLzoChunk() )
        {
            badHeader = true;
        }
    }

    public boolean isBadHeader()
    {
        return badHeader;
    }

    public void writeHeader( OutputStream outputStream )
        throws IOException
    {
        outputStream.write( header );
    }

    public byte getFlags()
    {
        return header[1];
    }

    public void setFlags( int flags )
    {
        header[1] = (byte) flags;
    }

    public boolean hasFlags( int flags )
    {
        return ( getFlags() & flags ) > 0;
    }

    public boolean isRencodeEnabled()
    {
        return hasFlags( FLAGS_RENCODE );
    }

    public byte getCompression()
    {
        return header[2];
    }

    public byte getCompressionLevel()
    {
        return (byte) ( getCompression() & 0xFF );
    }

    public boolean isBrotliChunk()
    {
        return ( getCompression() & BROTLI_FLAG ) > 0;
    }

    public boolean isLz4Chunk()
    {
        return ( getCompression() & LZ4_FLAG ) > 0;
    }

    public boolean isLzoChunk()
    {
        return ( getCompression() & LZO_FLAG ) > 0;
    }

    public boolean isDataCompressed()
    {
        return getCompressionLevel() > 0;
    }

    public int getPacketIndex()
    {
        return header[3];
    }

    public void setPacketIndex( int packetIndex )
    {
        header[3] = (byte) packetIndex;
    }

    public int getPacketSize()
    {
        return ( header[4] & 0xFF ) << 24 | ( header[5] & 0xFF ) << 16 | ( header[6] & 0xFF ) << 8 | ( header[7] & 0xFF );
    }

    public void setPacketSize( int packetSize )
    {
        header[4] = (byte) ( ( packetSize >>> 24 ) & 0xFF );
        header[5] = (byte) ( ( packetSize >>> 16 ) & 0xFF );
        header[6] = (byte) ( ( packetSize >>> 8 ) & 0xFF );
        header[7] = (byte) ( packetSize & 0xFF );
    }

    @Override
    public String toString()
    {
        return ( getClass().getSimpleName() + Arrays.toString( header ) );
    }

}
