/**
 * Copyright (C) 2020-2022 Dušan Vejnović <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dussan.vaadin.dvirtual.xpra.packets.server;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import org.apache.commons.lang3.StringUtils;
import org.dussan.vaadin.dvirtual.xpra.packets.PacketType;
import org.dussan.vaadin.dvirtual.xpra.packets.ServerPacket;
import org.dussan.vaadin.dvirtual.xpra.packets.base.ClientCapabilities;
import org.dussan.vaadin.dvirtual.xpra.packets.base.WindowType;

public class ServerWindowMetadataPacket
    extends ServerPacket
{

    private static final long serialVersionUID = 7840403270556752019L;

    private static final String CLASS_INSTANCE = ClientCapabilities.Metadata.CLASS_INSTANCE;

    private static final String ICON = "icon";

    private static final String MAXIMIZED = ClientCapabilities.Metadata.MAXIMIZED;

    private static final String MINIMUM_SIZE = "minimum-size";

    private static final String SIZE_CONSTRAINT = "size-constraints";

    private static final String TITLE = ClientCapabilities.Metadata.TITLE;

    private static final String TRANSIENT_FOR = ClientCapabilities.Metadata.TRANSIENT_FOR;

    private static final String WINDOW_TYPE = ClientCapabilities.Metadata.WINDOW_TYPE;

    public static final String OVERRIDE_REDIRECT = ClientCapabilities.Metadata.OVERRIDE_DIRECT;

    public static final int NO_PARENT = -1;

    private int windowId = 0;

    private transient Map<String, Object> metaData = null;

    public ServerWindowMetadataPacket()
    {
        super( PacketType.Server.WINDOW_METADATA.toString() );
        metaData = new HashMap<>();
    }

    public ServerWindowMetadataPacket( int windowId, Map<String, Object> metaData )
    {
        super( PacketType.Server.WINDOW_METADATA.toString() );
        this.windowId = windowId;
        this.metaData = metaData;
    }

    public void put( String key, Object value )
    {
        metaData.put( key, value );
    }

    public String getString( String key )
    {
        return getString( metaData.get( key ) );
    }

    public boolean getBoolean( String key )
    {
        return getBoolean( metaData.get( key ) );
    }

    public Integer getInt( String key )
    {
        return getInt( metaData.get( key ) );
    }

    @SuppressWarnings( "unchecked" )
    public ServerWindowIconPacket getIcon()
    {
        List<Object> packetData = (List<Object>) metaData.get( ICON );
        ServerWindowIconPacket icon = new ServerWindowIconPacket( windowId );
        if ( Objects.nonNull( packetData ) )
        {
            if ( !( packetData.get( 0 ) instanceof String ) )
            {
                packetData.add( windowId, 0 );
            }
            icon.unpackPacket( packetData.iterator() );
        }
        return icon;
    }

    public int getWindowId()
    {
        return windowId;
    }

    public int getParentId()
    {
        Integer value = null;
        if ( metaData.containsKey( TRANSIENT_FOR ) )
        {
            value = getInt( metaData.get( TRANSIENT_FOR ) );
        }
        return ( Objects.isNull( value ) ? NO_PARENT : value );
    }

    public String getTitle()
    {
        String value = StringUtils.EMPTY;
        if ( metaData.containsKey( TITLE ) )
        {
            value = getString( metaData.get( TITLE ) );
        }
        return value;
    }

    public String getApplicationName()
    {
        String value = StringUtils.EMPTY;
        if ( metaData.containsKey( CLASS_INSTANCE ) )
        {
            List<?> list = (List<?>) metaData.get( CLASS_INSTANCE );
            value = getString( list.get( 0 ) );
        }
        return value;
    }

    public WindowType getWindowType()
    {
        String value = StringUtils.EMPTY;
        if ( metaData.containsKey( WINDOW_TYPE ) )
        {
            List<?> list = (List<?>) metaData.get( WINDOW_TYPE );
            value = getString( list.get( 0 ) );
        }
        return WindowType.getType( value );
    }

    public int getMinWidth()
    {
        int value = 0;
        if ( metaData.containsKey( SIZE_CONSTRAINT ) )
        {
            Map<?, ?> map = (Map<?, ?>) metaData.get( SIZE_CONSTRAINT );
            if ( map.containsKey( MINIMUM_SIZE ) )
            {
                List<?> minSize = (ArrayList<?>) map.get( MINIMUM_SIZE );
                value = getInt( minSize.get( 0 ) );
            }
        }
        return value;
    }

    public int getMinHeight()
    {
        int value = 0;
        if ( metaData.containsKey( SIZE_CONSTRAINT ) )
        {
            Map<?, ?> map = (Map<?, ?>) metaData.get( SIZE_CONSTRAINT );
            if ( map.containsKey( MINIMUM_SIZE ) )
            {
                List<?> minSize = (ArrayList<?>) map.get( MINIMUM_SIZE );
                value = getInt( minSize.get( 1 ) );
            }
        }
        return value;
    }

    public boolean isMaximized()
    {
        boolean value = false;
        if ( metaData.containsKey( MAXIMIZED ) )
        {
            value = getBoolean( MAXIMIZED );
        }
        return value;
    }

    public boolean isOverrideRedirect()
    {
        boolean value = false;
        if ( metaData.containsKey( OVERRIDE_REDIRECT ) )
        {
            value = getBoolean( OVERRIDE_REDIRECT );
        }
        return value;
    }

    @SuppressWarnings( "unchecked" )
    @Override
    public void unpackPacket( Iterator<Object> packetData )
    {
        super.unpackPacket( packetData );
        windowId = getInt( packetData.next() );
        metaData.putAll( (Map<String, Object>) packetData.next() );
    }

}
